import numpy as np
from Elasticipy.tensors.second_order import SymmetricSecondOrderTensor


class StrainTensor(SymmetricSecondOrderTensor):
    """
    Class for manipulating symmetric strain tensors or arrays of symmetric strain tensors.

    """
    name = 'Strain tensor'
    voigt_map = [1, 1, 1, 2, 2, 2]

    def principal_strains(self):
        """
        Values of the principals strains.

        If the tensor array is of shape [m,n,...], the results will be of shape [m,n,...,3].

        Returns
        -------
        np.ndarray
            Principal strain values
        """
        return self.eigvals()

    def volumetric_strain(self):
        """
        Volumetric change (1st invariant of the strain tensor)

        Returns
        -------
        numpy.ndarray or float
            Volumetric change
        """
        return self.I1

    def eq_strain(self):
        """von Mises equivalent strain"""
        return np.sqrt(2/3 * self.ddot(self))

    def elastic_energy(self, stress, mode='pair'):
        """
        Compute the elastic energy.

        Parameters
        ----------
        stress : StressTensor
            Corresponding stress tensor
        mode : str, optional
            If 'pair' (default), the elastic energies are computed element-wise. Broadcasting rule applies.
            If 'cross', each cross-combination of stress and strain are considered.

        Returns
        -------
        float or numpy.ndarray
            Volumetric elastic energy
        """
        return 0.5 * self.ddot(stress, mode=mode)

    def draw_Mohr_circles(self):
        fig, ax = super().draw_Mohr_circles()
        ax.set_xlabel(ax.get_xlabel() + ' strain')
        ax.set_ylabel(ax.get_ylabel() + ' strain')
        return fig, ax


class StressTensor(SymmetricSecondOrderTensor):
    """
    Class for manipulating stress tensors or arrays of stress tensors.
    """
    name = 'Stress tensor'

    def principal_stresses(self):
        """
        Values of the principals stresses.

        If the tensor array is of shape [m,n,...], the results will be of shape [m,n,...,3].

        Returns
        -------
        np.ndarray
            Principal stresses
        """
        return self.eigvals()

    def vonMises(self):
        """
        von Mises equivalent stress.

        Returns
        -------
        np.ndarray or float
            von Mises equivalent stress

        See Also
        --------
        Tresca : Tresca equivalent stress
        """
        return np.sqrt(3 * self.J2)

    def Tresca(self):
        """
        Tresca(-Guest) equivalent stress.

        Returns
        -------
        np.ndarray or float
            Tresca equivalent stress

        See Also
        --------
        vonMises : von Mises equivalent stress
        """
        ps = self.principal_stresses()
        return ps[...,0] - ps[...,-1]

    def hydrostatic_pressure(self):
        """
        Hydrostatic pressure

        Returns
        -------
        np.ndarray or float

        See Also
        --------
        sphericalPart : spherical part of the stress
        """
        return -self.I1/3

    def elastic_energy(self, strain, mode='pair'):
        """
        Compute the elastic energy.

        Parameters
        ----------
        strain : StrainTensor
            Corresponding elastic strain tensor
        mode : str, optional
            If 'pair' (default), the elastic energies are computed element-wise. Broadcasting rule applies.
            If 'cross', each cross-combination of stress and strain are considered.

        Returns
        -------
        numpy.ndarray
            Volumetric elastic energy
        """
        return 0.5 * self.ddot(strain, mode=mode)

    def draw_Mohr_circles(self):
        fig, ax = super().draw_Mohr_circles()
        ax.set_xlabel(ax.get_xlabel() + ' stress')
        ax.set_ylabel(ax.get_ylabel() + ' stress')
        return fig, ax


from Elasticipy.tensors.elasticity import StiffnessTensor
from matplotlib import pyplot as plt

C = StiffnessTensor.orthorhombic(phase_name='forsterite',
                                C11=320, C12=68.2, C13=71.6,
                                C22=196.5, C23=76.8, C33=233.5, C44=64, C55=77, C66=78.7)
rho = 3.355

cp, cs_fast, cs_slow = C.wave_velocity(rho)
fig = plt.figure(figsize=(12,4))
cp.plot_as_pole_figure(subplot_args=(131,), title='p wave', fig=fig)
cs_fast.plot_as_pole_figure(subplot_args=(132,), title='s wave (fast)', fig=fig)
cs_slow.plot_as_pole_figure(subplot_args=(133,), title='s wave (slow)', fig=fig)
fig.show()


import numpy as np
from Elasticipy.tensors.stress_strain import StressTensor, StrainTensor
from Elasticipy.tensors.elasticity import StiffnessTensor
from scipy.spatial.transform import Rotation

# ======================================================
# Simple example of stress
# ======================================================
stress = StressTensor.shear([1,0,0],[0,1,0],1)
print(stress.vonMises())
print(stress.Tresca())

# ======================================================
# Simple example of strain
# ======================================================
strain = StrainTensor.shear([1,0,0],[0,1,0],1e-3)
print(strain.principal_strains())
print(strain.volumetric_strain())

# ======================================================
# Linear elasticity
# ======================================================
C = StiffnessTensor.cubic(phase_name='ferrite',C11=274, C12=175, C44=89)
print(C)

sigma = C * strain
print(sigma)

S = C.inv()
print(S)
print(S * sigma)

# ======================================================
# Multidimensional tensor arrays
# ======================================================
n_array = 10
sigma_xy = np.linspace(0, 100, n_array)
sigma = StressTensor.shear([1,0,0], [0,1,0],sigma_xy)
print(sigma[0])     # Check the initial value of the stress...
print(sigma[-1])    # ...and the final value.

eps = S * sigma
print(eps[0])     # Now check the initial value of strain...
print(eps[-1])    # ...and the final value.

energy = 0.5*sigma.ddot(eps)
print(energy)     # print the elastic energy

# ======================================================
# Apply random rotations
# ======================================================
n_ori = 1000
rotations = Rotation.random(n_ori)

eps_rotated = eps.rotate(rotations, mode='cross')
print(eps_rotated.shape)    # Just to check how it looks like

sigma_rotated = C * eps_rotated
print(sigma_rotated.shape)    # Check out the shape of the stresses

sigma = sigma_rotated * rotations.inv()         # Go back to initial frame
sigma_mean = sigma.mean(axis=1)     # Compute the mean over all orientations
print(sigma_mean[-1])

C_rotated = C * rotations
C_Voigt = C_rotated.Voigt_average() # Now check that the previous result is consistent with Voigt average
sigma_Voigt = C_Voigt * eps
print(sigma_Voigt[-1])

fig, ax = sigma_Voigt[-1].draw_Mohr_circles()
fig.show()




from Elasticipy.plasticity import JohnsonCook
from Elasticipy.tensors.stress_strain import StressTensor, StrainTensor
from Elasticipy.tensors.elasticity import StiffnessTensor
import numpy as np
from matplotlib import pyplot as plt


JC = JohnsonCook(A=363, B=792.7122, n=0.5756) # https://doi.org/10.1016/j.matpr.2020.05.213
C = StiffnessTensor.isotropic(E=210000, nu=0.27)

n_step = 100
sigma_max = 725
stress_mag = np.linspace(0, sigma_max, n_step)
stress = StressTensor.shear([1,0,0], [0,1,0],stress_mag)

elastic_strain = C.inv() * stress
plastic_strain = StrainTensor.zeros(n_step)
for i in range(2, n_step):
    strain_increment = JC.compute_strain_increment(stress[i])
    plastic_strain[i] = plastic_strain[i-1] + strain_increment


eps_xx = elastic_strain.C[0,1]+plastic_strain.C[0,1]
fig, ax = plt.subplots()
ax.plot(eps_xx, stress_mag, label='Stress-controlled')
ax.set_xlabel(r'$\varepsilon_{xy}$')
ax.set_ylabel('Shear stress (MPa)')

##
from scipy.optimize import minimize_scalar
stress = StressTensor.zeros(n_step)
plastic_strain = StrainTensor.zeros(n_step)
JC.reset_strain()
for i in range(2, n_step):
    def fun(tensile_stress):
        trial_stress = StressTensor.shear([1,0,0], [0,1,0], tensile_stress)
        trial_elastic_strain = C.inv() * trial_stress
        trial_strain_increment = JC.compute_strain_increment(trial_stress, apply_strain=False)
        trial_plastic_strain = plastic_strain[i - 1] + trial_strain_increment
        trial_elongation =  trial_plastic_strain.C[0,1] +  trial_elastic_strain.C[0,1]
        return (trial_elongation - eps_xx[i])**2
    s = minimize_scalar(fun)
    s0 = s.x
    stress.C[0,1][i] = s0
    strain_increment = JC.compute_strain_increment(stress[i])
    plastic_strain[i] = plastic_strain[i-1] + strain_increment

ax.plot(eps_xx, stress.C[0,1], label='Strain-controlled', linestyle='dotted')
ax.legend()


#
JC.reset_strain()
JC_tresca = JohnsonCook(A=363, B=792.7122, n=0.5756, criterion='Tresca')
stress_mag = np.linspace(0, 500, n_step)
stress = StressTensor.shear([1,0,0], [0,1,0],stress_mag)
models = (JC, JC_tresca)
labels = ('von Mises', 'Tresca')

elastic_strain = C.inv() * stress
fig, ax = plt.subplots()
for j, model in enumerate(models):
    plastic_strain = StrainTensor.zeros(n_step)
    for i in range(2, n_step):
        strain_increment = model.compute_strain_increment(stress[i])
        plastic_strain[i] = plastic_strain[i-1] + strain_increment
    eps_xy = elastic_strain.C[0,1]+plastic_strain.C[0,1]
    ax.plot(eps_xy, stress_mag, label=labels[j])
ax.set_xlabel(r'$\varepsilon_{xy}$')
ax.set_ylabel('Shear stress (MPa)')
ax.legend()
fig.show()


import unittest
from Elasticipy.tensors.fourth_order import FourthOrderTensor, SymmetricFourthOrderTensor
import numpy as np

from Elasticipy.tensors.second_order import SecondOrderTensor


class TestFourthOrderTensor(unittest.TestCase):
    def test_multidimensionalArrayTensors(self):
        m = 5
        a = np.random.random((m, 6, 6))
        T = FourthOrderTensor(a)
        np.testing.assert_array_almost_equal(a, T._matrix)
        T2 = FourthOrderTensor(T.full_tensor())
        np.testing.assert_array_almost_equal(a, T2._matrix)

    def test_nonsymmetry(self):
        a = np.random.random((3,3,3,3))
        with self.assertRaises(ValueError) as context:
            _ = FourthOrderTensor(a)
        self.assertEqual(str(context.exception), 'The input array does not have minor symmetry')
        T = FourthOrderTensor(a, force_minor_symmetry=True)
        Tfull = T.full_tensor()
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    for l in range(3):
                        b = 0.25 * (a[i,j,k,l] + a[j,i,k,l] + a[i,j,l,k] + a[j,i,l,k])
                        np.testing.assert_array_almost_equal(Tfull[i,j,k,l], b)

    def test_inversion(self):
        m = 5
        a = np.random.random((m, 6, 6))
        T = FourthOrderTensor(a)
        Tinv = T.inv()
        TTinv = Tinv.ddot(T)
        eye = FourthOrderTensor.identity(m)
        for i in range(m):
            np.testing.assert_array_almost_equal(TTinv[i].full_tensor(), eye[i].full_tensor())

    def test_mult(self):
        m, n, o = 5, 4, 3
        a = np.random.random((m,n,o,6,6))
        a = FourthOrderTensor(a)
        b = 5
        ab = a * b
        for i in range(m):
            for j in range(n):
                for k in range(o):
                    np.testing.assert_array_equal(ab[i,j,k]._matrix, a[i,j,k]._matrix * b)

        b = np.random.random((n,o))
        ab = a * b
        for i in range(m):
            for j in range(n):
                for k in range(o):
                    np.testing.assert_array_equal(ab[i,j,k]._matrix, a[i,j,k]._matrix * b[j,k])

    def test_zeros_setitem(self):
        m, n = 4, 5
        t = FourthOrderTensor.zeros()
        assert t.shape == ()
        assert np.all(t.full_tensor()==0.)

        t = FourthOrderTensor.zeros(n)
        assert t.shape == (n,)

        t = FourthOrderTensor.zeros((m,n))
        assert t.shape == (m, n)

        t[1,3] = np.ones((6,6))
        for i in range(m):
            for j in range(n):
                if (i == 1) and (j == 3):
                    assert np.all(t[i,j] == 1.)
                else:
                    assert np.all(t[i, j] == 0.)

        t0 = t == 0.
        t0_th = np.ones((m, n))
        t0_th[1,3] = 0.
        assert np.all(t0== t0_th)

    def test_div(self):
        m, n, o = 5, 4, 3
        a = np.random.random((m, n, o, 6, 6))
        a = FourthOrderTensor(a)
        a_div_a = a / a
        np.testing.assert_array_almost_equal(a_div_a.full_tensor(), FourthOrderTensor.identity(return_full_tensor=True, shape=(m,n,o)))

        half_a = a / 2
        np.testing.assert_array_almost_equal(half_a.full_tensor(), a.full_tensor()/2)

        b = SecondOrderTensor.rand(shape=(4,3))
        a_div_b = a / b
        np.testing.assert_array_almost_equal(a_div_b.matrix, (a * b.inv()).matrix)



class TestSymmetricFourthOrderTensor(unittest.TestCase):
    def test_inversion(self):
        m = 5
        a = np.random.random((m, 6, 6))
        T = SymmetricFourthOrderTensor(a, force_symmetries=True)
        Tinv = T.inv()
        TTinv = Tinv.ddot(T)
        eye = SymmetricFourthOrderTensor.identity(m)
        for i in range(m):
            np.testing.assert_array_almost_equal(TTinv[i].full_tensor(), eye[i].full_tensor())

if __name__ == '__main__':
    unittest.main()



import unittest
from Elasticipy.tensors.thermal_expansion import ThermalExpansionTensor as ThEx
from scipy.spatial.transform import Rotation
import numpy as np

coeff = np.array([[11, 12, 13], [12, 22, 23], [13, 23, 33]])
alpha = ThEx(coeff)

class TestThermalExpansion(unittest.TestCase):
    def test_transverse(self):
        alpha = ThEx.transverse_isotropic(alpha_11=11, alpha_33=33)
        np.testing.assert_array_equal(alpha.matrix, np.diag([11, 11, 33]))

    def test_orthotropic(self):
        alpha_orthotropic = ThEx.orthotropic(11, 22, 33)
        alpha_orthorhombic= ThEx.orthorhombic(11, 22, 33)
        matrix = np.diag([11., 22., 33.])
        np.testing.assert_array_equal(alpha_orthotropic.matrix, matrix)
        np.testing.assert_array_equal(alpha_orthorhombic.matrix, matrix)

    def test_triclinic(self):
        alpha = ThEx.triclinic(alpha_11=11, alpha_12=12, alpha_13=13, alpha_22=22, alpha_23=23, alpha_33=33)
        matrix = np.array([[11., 12., 13.],[12., 22., 23.],[13., 23., 33.]])
        np.testing.assert_array_equal(alpha.matrix, matrix)

    def test_monoclinic(self):
        # Check monoclinic with X-Y mirror plane
        alpha = ThEx.monoclinic(alpha_11=11, alpha_22=22, alpha_33=33, alpha_12=12)
        np.testing.assert_array_equal(alpha.matrix, np.array([[11, 12, 0],[12, 22, 0],[0, 0, 33]], dtype=np.float64))

        # Check monoclinic with X-Z mirror plane
        alpha = ThEx.monoclinic(alpha_11=11, alpha_22=22, alpha_33=33, alpha_13=13)
        np.testing.assert_array_equal(alpha.matrix, np.array([[11, 0, 13],[0, 22, 0],[13, 0, 33]], dtype=np.float64))

        # Check error if alpha_12 and alpha_13 are provided
        with self.assertRaises(ValueError) as context:
            ThEx.monoclinic(alpha_11=11, alpha_22=22, alpha_33=33, alpha_12=12, alpha_13=13)
        self.assertEqual(str(context.exception), 'alpha_13 and alpha_12 cannot be used together.')

        # Check error if none of alpha_12 and alpha_13 are provided
        with self.assertRaises(ValueError) as context:
            ThEx.monoclinic(alpha_11=11, alpha_22=22, alpha_33=33)
        self.assertEqual(str(context.exception), 'Either alpha_13 or alpha_12 must be provided.')

    def test_volumetric(self):
        matrix = np.random.random((3,3))
        matrix = matrix + matrix.T
        alpha = ThEx(matrix)
        assert np.trace(matrix) == alpha.volumetric_coefficient

    def test_matmul_rotations(self):
        m = 50
        rotations = Rotation.random(m)
        alphas = alpha.matmul(rotations)
        for i in range(m):
            rot_mat = np.matmul(rotations[i].as_matrix().T, np.matmul(alpha.matrix, rotations[i].as_matrix()))
            np.testing.assert_almost_equal(rot_mat, alphas[i].matrix)

    def test_mul(self):
        n=50
        rotations = Rotation.random(n)
        alphas = alpha * rotations
        temp = np.linspace(0,10,n)
        eps = alphas * temp
        assert eps.shape == (n,)
        for i in range(n):
            rot_mat = rotations[i].as_matrix()
            rotated_strain_matrix = np.matmul(rot_mat.T, np.matmul(coeff, rot_mat)) * temp[i]
            np.testing.assert_almost_equal(eps[i].matrix, rotated_strain_matrix)

    def test_matmul_array(self):
        m, n= 50, 100
        rotations = Rotation.random(m)
        alphas = alpha * rotations
        temp = np.linspace(0,10,n)
        eps = alphas.matmul(temp)
        assert eps.shape == (m, n)
        for i in range(m):
            rot_mat = rotations[i].as_matrix()
            rotated_tensor_matrix = np.matmul(rot_mat.T, np.matmul(coeff, rot_mat))
            for j in range(n):
                np.testing.assert_almost_equal(eps[i,j].matrix, rotated_tensor_matrix * temp[j])

    def test_constructor_isotropic(self):
        coeff = 23e-6
        alpha = ThEx.isotropic(coeff)
        temp = 25
        eps = alpha * temp
        np.testing.assert_almost_equal(eps.matrix, np.eye(3) * coeff * temp)

    def test_apply_temperature(self):
        m, n = 50, 100
        rotations = Rotation.random(m)
        alphas = alpha * rotations
        T = np.arange(0, m)
        strain = alphas.apply_temperature(T)
        for i in range(m):
            np.testing.assert_array_equal(strain[i].matrix, alphas[i].matrix * T[i])
        T = np.arange(0, n)
        strain = alphas.apply_temperature(T, mode='cross')
        for i in range(m):
            for j in range(n):
                np.testing.assert_array_equal(strain[i,j].matrix, alphas[i].matrix * T[j])

    def test_deprecated_path(self):
        expected_warn = ("The module 'Elasticipy.ThermalExpansion' is deprecated and will be removed in a future "
                            "release. Please use 'Elasticipy.tensors.thermal_expansion' instead.")
        with self.assertWarns(DeprecationWarning) as context:
            from Elasticipy.ThermalExpansion import ThermalExpansionTensor
        self.assertEqual(str(context.warning), expected_warn)

if __name__ == '__main__':
    unittest.main()




from Elasticipy.tensors.elasticity import StiffnessTensor
import numpy as np
from matplotlib import pyplot as plt


C_ferrite = StiffnessTensor.cubic(C11=242, C12=146, C44=116)
C_austenite = StiffnessTensor.cubic(C11=204, C12=137, C44=126)

n_values = 100
fraction_austenite = np.linspace(0, 1, n_values)

fig, ax = plt.subplots()
colors = plt.cm.tab10.colors    # Simple hack to ensure that both E and G will have the same color for a given method
color_index = 0
for method in ('Voigt', 'Hill', 'Reuss'):
    C_aust_avg = C_austenite.average(method)    # Equivalent to C_austenite.{method}_average()
    C_ferr_avg = C_ferrite.average(method)
    E_mean = np.zeros(n_values)
    G_mean = np.zeros(n_values)
    for i in range(n_values):
        f1 = fraction_austenite[i]
        f2 = 1 - f1
        C_tot_mean = StiffnessTensor.weighted_average((C_aust_avg, C_ferr_avg), (f1, f2), method)
        E_mean[i] = C_tot_mean.Young_modulus.mean()
        G_mean[i] = C_tot_mean.shear_modulus.mean()
    ax.plot(fraction_austenite, E_mean, label='E ({})'.format(method), color=colors[color_index])
    ax.plot(fraction_austenite, G_mean, label='G ({})'.format(method), color=colors[color_index], linestyle='--')
    color_index = color_index + 1

ax.legend()
ax.set_xlim([0, 1])
ax.set_xlabel('Fraction of Austenite')
ax.set_ylabel('Young/Shear Modulus (GPa)')
fig.show()






from Elasticipy.tensors.stress_strain import StrainTensor
from Elasticipy.tensors.elasticity import StiffnessTensor
from matplotlib import pyplot as plt
import matplotlib as mpl
mpl.use('Qt5Agg')   # Ensure interactive plot
import numpy as np
import time

C = StiffnessTensor.transverse_isotropic(Ex=200, Ez=300, nu_yx=0.2, nu_zx=0.3, Gxz=80)
## Elasticipy's stuff
exp = np.arange(0, 8)
n = 10**exp
t_stress_elast=[]
t_stress_mg = []
t_vm_elast=[]
t_vm_mg = []

for ni in n:
    strain = StrainTensor.rand(shape=(ni,))
    start_time = time.perf_counter()
    stress = C*strain
    t_stress_elast.append(time.perf_counter() - start_time)
    start_time = time.perf_counter()
    vm_elast = stress.vonMises()
    t_vm_elast.append(time.perf_counter() - start_time)

    ## Pymatgen's stuff
    Cmg = C.to_pymatgen()
    strain_mg = strain.to_pymatgen()
    start_time = time.perf_counter()
    stress = [Cmg.calculate_stress(strain_mg[i]) for i in range(ni)]
    t_stress_mg.append(time.perf_counter() - start_time)
    start_time = time.perf_counter()
    vm_mg = [stress[i].von_mises for i in range(ni)]
    t_vm_mg.append(time.perf_counter() - start_time)


fig, ax = plt.subplots()
ax.plot(n, t_stress_mg, label="Generalized Hooke's law (Pymatgen)", marker="s")
ax.plot(n, t_stress_elast, label="Generalized Hooke's law (Elasticipy)", marker="o")
ax.plot(n, t_vm_mg, label='von Mises eq. stress (Pymatgen)', linestyle='dotted', marker="s")
ax.plot(n, t_vm_elast, label='von Mises eq. stress  (Elasticipy)', linestyle='dotted', marker="o")
plt.legend()
plt.xscale('log')
plt.yscale('log')
ax.set_xlabel('Number of tensors')
ax.set_ylabel('CPU time (s)')
ax.set_xlim((1, max(n)))
fig.tight_layout()
plt.show()
fig.savefig('../JOSS/ElasticipyVSpymatgen.png', dpi=300)


import unittest
from pytest import approx
import numpy as np
from scipy.spatial.transform import Rotation

from Elasticipy.tensors.elasticity import StiffnessTensor
import Elasticipy.tensors.stress_strain as Tensors
from Elasticipy.tensors.second_order import SecondOrderTensor, SymmetricSecondOrderTensor, SkewSymmetricSecondOrderTensor
from Elasticipy.tensors.stress_strain import StrainTensor, StressTensor
from pymatgen.analysis.elasticity import Strain as mgStrain, Stress as mgStress
from orix.quaternion import Rotation as OrixRot

Cmat = [[231, 127, 104, 0, -18, 0],
        [127, 240, 131, 0, 1, 0],
        [104, 131, 175, 0, -3, 0],
        [0, 0, 0, 81, 0, 3],
        [-18, 1, -3, 0, 11, 0],
        [0, 0, 0, 3, 0, 85]]
C = StiffnessTensor(Cmat)


class TestStressStrainTensors(unittest.TestCase):
    def test_mult_by_stiffness(self):
        """
        Test Stiffness/Strain tensors product C*eps (which stands for C:E)
        """
        tensile_dir = [1, 0, 0]
        stress = Tensors.StressTensor([[1, 0, 0],
                                       [0, 0, 0],
                                       [0, 0, 0]])
        strain = C.inv()*stress
        eps_xx = strain.C[0, 0]
        eps_yy = strain.C[1, 1]
        eps_zz = strain.C[2, 2]
        E = C.Young_modulus.eval(tensile_dir)
        nu_y = C.Poisson_ratio.eval(tensile_dir, [0, 1, 0])
        nu_z = C.Poisson_ratio.eval(tensile_dir, [0, 0, 1])
        assert eps_xx == approx(1/E)
        assert eps_yy == approx(-nu_y / E)
        assert eps_zz == approx(-nu_z / E)

    def test_rotate_tensor(self, n_oris=100):
        """
        Test the rotation of a tensor

        Parameters
        ----------
        n_oris : int
            Number of random orientations to use
        """
        random_tensor = SecondOrderTensor(np.random.random((3, 3)))
        random_oris = Rotation.random(n_oris)
        eps_rotated = random_tensor * random_oris
        for i in range(n_oris):
            rot_mat = random_oris[i].as_matrix()
            eps_matrix_th = np.matmul(np.matmul(rot_mat.T, random_tensor.matrix), rot_mat)
            np.testing.assert_almost_equal(eps_matrix_th, eps_rotated[i].matrix)

    def test_transpose_array(self):
        """
        Test transposing a tensor array
        """
        shape = (1, 2, 3)
        random_matrix = np.random.random(shape + (3, 3))
        random_tensor = SecondOrderTensor(random_matrix)
        transposed_tensor = random_tensor.T
        for i in range(shape[0]):
            for j in range(shape[1]):
                for k in range(shape[2]):
                    np.testing.assert_array_equal(random_matrix[i, j, k], transposed_tensor[k, j, i].matrix)

        # Check that transposing a single tensor has no effect
        a = SecondOrderTensor.rand()
        assert a == a.transpose_array()

    def test_transpose_tensor(self):
        """
        Test transposing a tensor array
        """
        shape = (2, 3, 4)
        tensor = SecondOrderTensor(np.random.random(shape + (3, 3)))
        tensor_transposed = tensor.transpose_tensor()
        for i in range(shape[0]):
            for j in range(shape[1]):
                for k in range(shape[2]):
                    np.testing.assert_array_equal(tensor_transposed[i, j, k].matrix, tensor[i, j, k].matrix.T)


    def test_mul(self):
        """
        Test the element-wise product of tensors.
        """
        # First, multiply a SymmetricSecondOrderTensor with another one, and expect a matrix product between each
        # sliced matrix
        shape = (4, 5)
        shape = shape + (3, 3)
        matrix1 = np.random.random(shape)
        matrix2 = np.random.random(shape)
        tensor_prod = SecondOrderTensor(matrix1) * SecondOrderTensor(matrix2)
        for i in range(shape[0]):
            for j in range(shape[1]):
                mat_prod = np.matmul(matrix1[i, j], matrix2[i, j])
                np.testing.assert_array_almost_equal(tensor_prod[i, j].matrix, mat_prod)

        # Now, multiply a SymmetricSecondOrderTensor with an array of the same shape, and expect an element-wise
        # multiplication between the sliced matrix of the tensor and the values of the array
        t = SecondOrderTensor(matrix1)
        random_array = np.random.random(t.shape)
        tensor_prod = t * random_array
        for i in range(shape[0]):
            for j in range(shape[1]):
                matrix = tensor_prod[i, j].matrix
                np.testing.assert_array_equal(matrix, t.matrix[i,j,:] * random_array[i,j])

    def test_truediv(self):
        a = Tensors.StressTensor.rand((5,6))
        adiv = a/2
        np.testing.assert_array_equal(adiv.matrix, a.matrix / 2)
        with self.assertRaises(NotImplementedError) as context:
            _ = a / a
        self.assertEqual(str(context.exception), 'Tensors can only be divided by scalar values or by arrays '
                                                 'of the same shape.')

        mag = np.linspace(1,2)
        b = Tensors.StrainTensor.tensile([1,0,0], mag)
        bdiv = b / mag
        np.testing.assert_array_almost_equal(bdiv.matrix, np.tile(np.diag([1.,0,0]),(50,1,1)))

    def test_matmul(self, length1=3, length2=4):
        """
        Test the matrix-like product of tensor arrays

        Parameters
        ----------
        length1 : int
            Length of the first array
        length2 : int
            Length of the second array
        """
        matrix1 = np.random.random((length1, 3, 3))
        matrix2 = np.random.random((length2, 3, 3))
        rand_tensor1 = SecondOrderTensor(matrix1)
        rand_tensor2 = SecondOrderTensor(matrix2)
        cross_prod_tensor = rand_tensor1.dot(rand_tensor2, mode='cross')
        for i in range(0, length1):
            for j in range(0, length2):
                mat_prod = np.matmul(matrix1[i], matrix2[j])
                np.testing.assert_array_almost_equal(cross_prod_tensor[i, j].matrix, mat_prod)

    def test_matmul_rotation(self):
        m, n = 5, 100
        random_tensor = SecondOrderTensor(np.random.random((m,) + (3, 3)))
        random_oris = Rotation.random(n)
        array = random_tensor.rotate(random_oris, mode='cross')
        assert array.shape == (m, n)
        for i in range(m):
            for j in range(n):
                rot_mat = random_oris[j].as_matrix()
                matrix = np.matmul(np.matmul(rot_mat.T, random_tensor[i].matrix), rot_mat)
                np.testing.assert_almost_equal(matrix, array[i,j].matrix)

    def test_statistics(self):
        """
        Test the std, min and max functions for tensor arrays.
        """
        shape = (5, 4, 3, 2)
        matrix = np.random.random(shape + (3, 3))
        tensor = SecondOrderTensor(matrix)
        mini = tensor.min()
        maxi = tensor.max()
        std = tensor.std()
        # First, check T.std()
        for i in range(0, 3):
            for j in range(0, 3):
                Cij = matrix[..., i, j].flatten()
                assert np.std(Cij) == approx(std.C[i, j])
                assert np.min(Cij) == approx(mini.C[i, j])
                assert np.max(Cij) == approx(maxi.C[i, j])
        # Then, check T.std(axis=...)
        for i in range(0, len(shape)):
            np.testing.assert_array_equal(tensor.std(axis=i).matrix, np.std(matrix, axis=i))
            np.testing.assert_array_equal(tensor.min(axis=i).matrix, np.min(matrix, axis=i))
            np.testing.assert_array_equal(tensor.max(axis=i).matrix, np.max(matrix, axis=i))

        # Now, check for single value tensors
        tensor = tensor[0,0,0,0]
        np.testing.assert_array_equal(tensor.min().matrix, tensor.matrix)
        np.testing.assert_array_equal(tensor.max().matrix, tensor.matrix)
        np.testing.assert_array_equal(tensor.std().matrix, np.zeros((3, 3)))

    def test_ddot(self):
        """
        Test the ddot method.
        """
        shape = (4, 3, 2)
        tens1 = SecondOrderTensor.rand(shape)
        tens2 = SecondOrderTensor.rand(shape[1:])   # Force tens2 to have a different shape, just to check broadcasting
        ddot = tens1.ddot(tens2)
        for i in range(0, shape[0]):
            for j in range(0, shape[1]):
                for k in range(0, shape[2]):
                    ddot_th = np.trace(np.matmul(tens1.matrix[i,j,k].T, tens2.matrix[j,k]))
                    assert ddot_th == approx(ddot[i, j, k])

    def test_vonMises_Tresca(self):
        """
        Check that the Tresca and von Mises methods work well for simple tension, simple shear and hydrostatic
        pressure.
        """
        matrix = np.zeros((3, 3, 3))
        matrix[0, 0, 0] = 1  # Simple tension
        matrix[1, 1, 0] = matrix[1, 0, 1] = 1  # Simple shear
        matrix[2, np.arange(3), np.arange(3)] = -1  # Hydrostatic pressure
        stress = Tensors.StressTensor(matrix)

        vM_stress = stress.vonMises()
        vm_th = np.array([1, np.sqrt(3), 0.0])
        np.testing.assert_array_equal(vM_stress, vm_th)

        Tresca_stress = stress.Tresca()
        Tresca_th = np.array([1, 2, 0.0])
        np.testing.assert_array_equal(Tresca_stress, Tresca_th)

    def test_rotation_stiffness(self, ):
        """
        Check that the two ways to compute stress from a rotated stiffness tensor are consistent.
        """
        n_strain = 50
        n_ori = 100
        matrix = np.random.random((n_strain, 3, 3))
        eps = Tensors.StrainTensor(matrix, force_symmetry=True)
        ori = Rotation.random(n_ori)
        C_rotated = C * ori
        sigma = C_rotated.ddot(eps, mode='cross')

        # Rotate stress and stress by their own
        eps_rot = eps.rotate(ori, mode='cross')
        sigma_rot2 = C * eps_rot
        sigma2 = sigma_rot2 * ori.inv()
        np.testing.assert_almost_equal(sigma.matrix, sigma2.transpose_array().matrix)

    def test_multidimensional_tensors(self, ):
        """
        Check that the shape of (C * rotations) * eps is (m, p, r, ...) if rotation.shape=(p, q,...) and
        len(rotations)=m.
        """
        shape_strain = (5, 4, 3)
        n_ori = 100
        strain = Tensors.StrainTensor.ones(shape_strain)
        ori = Rotation.random(n_ori)
        C_rotated = C * ori
        stress = C_rotated.ddot(strain, mode='cross')
        self.assertEqual(stress.shape, (n_ori,) + shape_strain)
        for i in range(5):
            for j in range(4):
                for k in range(3):
                    assert np.all(stress[:,i,j,k] == C_rotated * strain[i,j,k])

    def test_Voigt_notation_strain(self):
        """
        Check that the strain tensor can be reconstructed from Voigt vectors.
        Returns
        """
        a = np.random.random((3, 4, 6))
        strain = Tensors.StrainTensor.from_Voigt(a)
        for i in range(0, 3):
            for j in range(0, 4):
                for k in range(0, 6):
                    if k<3:
                        assert a[i,j,k] == strain[i,j].C[k,k]
                    elif k==3:
                        assert a[i,j,k] == 2*strain[i,j].C[1,2]
                    elif k==4:
                        assert a[i,j,k] == 2*strain[i,j].C[0,2]
                    else:
                        assert a[i,j,k] == 2*strain[i,j].C[0,1]

    def test_Voigt_notation_stress(self):
        """
        Check that the stress tensor can be reconstructed from Voigt vectors.
        """
        a = np.random.random((3, 4 , 6))
        stress = Tensors.StressTensor.from_Voigt(a)
        for i in range(0, 3):
            for j in range(0, 4):
                for k in range(0, 6):
                    if k<3:
                        assert a[i,j,k] == stress[i,j].C[k,k]
                    elif k==3:
                        assert a[i,j,k] == stress[i,j].C[1,2]
                    elif k==4:
                        assert a[i,j,k] == stress[i,j].C[0,2]
                    else:
                        assert a[i,j,k] == stress[i,j].C[0,1]

    def test_tensile_stress(self):
        """Check that a stress tensor can be defined for tensile state"""
        n = 10
        sigma_11 = np.linspace(0,1, n)
        stress = Tensors.StressTensor.tensile([1,0,0], sigma_11)
        for i in range(0, n):
            stress_i = np.diag([sigma_11[i], 0, 0])
            np.testing.assert_array_equal(stress[i].matrix, stress_i)

    def test_shear_stress(self):
        """Check that a stress tensor can be defined for shear state"""
        n = 10
        sigma_12 = np.linspace(0,1, n)
        stress = Tensors.StressTensor.shear([1,0,0], [0,1,0], sigma_12)
        for i in range(0, n):
            stress_i = np.zeros((3,3))
            stress_i[0,1] = stress_i[1,0] = sigma_12[i]
            np.testing.assert_array_equal(stress[i].matrix, stress_i)

        # Now check if error is thrown if the two vectors are not orthogonal
        with self.assertRaises(ValueError) as context:
            Tensors.StrainTensor.shear([1,0,0], [1,1,0], 0.1)
        self.assertEqual(str(context.exception), 'u and v must be orthogonal')

    def test_set_item(self):
        """Check setting a tensor in a tensor array"""
        stress = Tensors.StressTensor.zeros((3, 3))
        stress[0,0] = np.ones(3)
        matrix = np.zeros((3, 3, 3, 3))
        matrix[0, 0, :, :] = 1
        np.testing.assert_array_equal(stress.matrix, matrix)

    def test_add_sub_mult_strain(self):
        """Check addition, subtraction and float multiplication of tensors"""
        shape = (3,3,3)
        a = Tensors.StrainTensor.ones(shape)
        b = 2 * Tensors.StrainTensor.ones(shape)
        c = 3 * Tensors.StrainTensor.ones(shape)
        d = a + b - c + 5 - 5 - a
        np.testing.assert_array_equal(d.matrix, -np.ones(shape + (3,3)))

    def test_flatten(self):
        """Check flattening of a tensor array"""
        shape = (3,3,3)
        matrix = np.random.random(shape + (3,3))
        a = SecondOrderTensor(matrix)
        a_flat = a.flatten()

        # Fist, check that the shapes are consistent
        assert a_flat.shape == np.prod(shape)

        # Then, check out each element
        for p in range(0, np.prod(shape)):
            i, j, k = np.unravel_index(p, shape)
            np.testing.assert_array_equal(a_flat[p].matrix, a[i,j,k].matrix)

    def test_symmetric_skew_parts(self):
        """Check the values returned by the symmetric and skew parts of a tensor"""
        shape = (2,3,4)
        a = SecondOrderTensor(np.random.random(shape + (3,3)))
        a_symm = a.symmetric_part()
        a_skew = a.skew_part()
        for i in range(0, shape[0]):
            for j in range(0, shape[1]):
                for k in range(0, shape[2]):
                    matrix = a[i,j,k].matrix
                    np.testing.assert_array_equal(2 * a_symm[i, j, k].matrix, matrix + matrix.T)
                    np.testing.assert_array_equal(2 * a_skew[i, j, k].matrix, matrix - matrix.T)

    def test_equality(self):
        """Test the == operator"""
        # Test equality for two tensors of the same shape
        shape = (3,4,5)
        a = SecondOrderTensor(np.random.random(shape + (3,3)))
        b = SecondOrderTensor(np.random.random(shape + (3,3)))
        a[0,1,2] = b[0,1,2]
        is_equal = a == b
        assert is_equal.shape == shape
        for i in range(0, shape[0]):
            for j in range(0, shape[1]):
                for k in range(0, shape[2]):
                    assert is_equal[i ,j, k] == np.all(a[i, j, k,:,:].matrix == b[i, j, k, :, :].matrix, axis=(-2, -1))

        # Test equality for an array of tensors, and a single tensor
        c = a[2, 1, 0]
        is_equal = a == c
        assert is_equal.shape == shape
        for i in range(0, shape[0]):
            for j in range(0, shape[1]):
                for k in range(0, shape[2]):
                    assert is_equal[i ,j, k] == np.all(a[i, j, k,:,:].matrix == c.matrix, axis=(-2, -1))

        # Now test inconsistent shapes
        shape2 = (3,4,5,6)
        d = SecondOrderTensor(np.random.random(shape2 + (3, 3)))
        expected_error = 'The value to compare must be an array of shape {} or {}'.format(shape, shape + (3,3))
        with self.assertRaises(ValueError) as context:
            _ = a == d
        self.assertEqual(str(context.exception), expected_error)

    def test_divergence(self):
        """Test the divergence operator"""
        spacing = [0.1, 0.2, 0.3]
        x = np.arange(0, 1, spacing[0])
        y = np.arange(0, 1, spacing[1])
        z = np.arange(0, 1, spacing[2])
        x, y, z = np.meshgrid(x, y, z, indexing='ij')
        shape = x.shape
        a_11, b_11, c_11 = 2., 3., -1.
        a_12, b_12, c_12 = 4, -1., 2.
        a_23, b_23, c_23 = 1, -1, 3.

        s = SecondOrderTensor.zeros(shape=shape)
        s.C[0, 0] = a_11 * x + b_11 * y + c_11 * z
        s.C[0, 1] = a_12 * x + b_12 * y + c_12 * z
        s.C[1, 2] = a_23 * x + b_23 * y + c_23 * z
        div_space = s.div(spacing=spacing)
        div_nonspaced = s.div()
        div_uniaxial = s.div(axes=0)
        div_biaxial = s.div(axes=(0,1))
        expected_div_spaced = [a_11 + b_12, c_23, 0]
        expected_div_nonspaced = [a_11 * spacing[0] + b_12 * spacing[1], c_23 * spacing[2], 0]
        expected_div_uniaxial = [a_11 * spacing[0], 0, 0]
        expected_div_biaxial = [a_11 * spacing[0] + b_12 * spacing[1], 0, 0]
        for i in range(0, shape[0]):
            for j in range(0, shape[1]):
                for k in range(0, shape[2]):
                    np.testing.assert_almost_equal(div_space[i, j, k], expected_div_spaced)
                    np.testing.assert_almost_equal(div_nonspaced[i, j, k], expected_div_nonspaced)
                    np.testing.assert_almost_equal(div_uniaxial[i, j, k], expected_div_uniaxial)
                    np.testing.assert_almost_equal(div_biaxial[i, j, k], expected_div_biaxial)


    def test_reshape(self):
        """Test reshaping and 'un-reshaping' an array of tensors"""
        init_shape = (6, 4, 5)
        new_shape = (30, 4)
        a = np.random.random(init_shape + (3, 3))
        t = SecondOrderTensor(a)
        t_reshaped = t.reshape(new_shape)
        assert t_reshaped.shape == new_shape
        t_reshaped_back = t_reshaped.reshape(init_shape)
        for i in range(0, init_shape[0]):
            for j in range(0, init_shape[1]):
                for k in range(0, init_shape[2]):
                    old_mat = a[i,j,k]
                    new_mat = t_reshaped_back[i,j,k].matrix
                    np.testing.assert_array_equal(old_mat, new_mat)

    def test_save_load_tensor(self):
        """Test save and load data to/from file"""

        # First, check with consistent shape
        a = np.random.random((5, 4, 3, 3))
        t = SecondOrderTensor(a)
        file_name = 'test_save.npy'
        t.save(file_name)
        t2 = t.load_from_npy(file_name)
        np.testing.assert_array_equal(t.matrix, t2.matrix)

        # Now try with inconsistent shape
        np.save(file_name, a[...,0])
        expected_error = 'The shape of the array to load must be (...,3,3).'
        with self.assertRaises(ValueError) as context:
            _ = SecondOrderTensor.load_from_npy(file_name)
        self.assertEqual(str(context.exception), expected_error)

    def test_save_load_csv(self):
        """Test saving and reading a CSV file"""
        # Check that when exporting/importing, we get the same tensor
        a = np.random.random((10, 3, 3))
        t = SecondOrderTensor(a)
        file_name = 'test_textfile.txt'
        t.save_as_txt(file_name)
        t2 = SecondOrderTensor.load_from_txt(file_name)
        np.testing.assert_array_almost_equal(t.matrix, t2.matrix)

        # Try with non-flatten tensor
        a = np.random.random((5, 3, 3, 3))
        t = SecondOrderTensor(a)
        expected_error = 'The array must be flatten before getting dumped to text file.'
        with self.assertRaises(ValueError) as context:
            t.save_as_txt(file_name)
        self.assertEqual(str(context.exception), expected_error)

        # Now try with a symmetric tensor
        a = np.random.random((10, 3, 3))
        a = a + np.swapaxes(a, -1, -2)
        t = SymmetricSecondOrderTensor(a)
        file_name = 'test_symmetric_textfile.txt'
        t.save_as_txt(file_name)
        t2 = SymmetricSecondOrderTensor.load_from_txt(file_name)
        np.testing.assert_array_almost_equal(t.matrix, t2.matrix)


    def test_symmetric_tensor_constructor(self):
        """Test constructor for symmetric second Order tensors"""

        # When a symmetric matrix is passed to the constructor
        mat = np.random.random((3,3))
        sym_mat = mat + mat.T
        t = SymmetricSecondOrderTensor(sym_mat)
        np.testing.assert_array_equal(t.matrix, sym_mat)

        # When an upper-diagonal matrix is passed to the constructor
        upper_mat = sym_mat
        upper_mat[np.tril_indices(3, -1)] = 0 # Set lower part to zero
        t2 = SymmetricSecondOrderTensor(upper_mat)
        assert t == t2

        # Expect error in any other case
        expected_error = ('The input array must be either slices of symmetric matrices, of slices of upper-diagonal '
                          'matrices.')
        with self.assertRaises(ValueError) as context:
            _ = SymmetricSecondOrderTensor(mat)
        self.assertEqual(str(context.exception), expected_error)

    def test_skew_symmetric_tensor_constructor(self):
        """Test constructor for symmetric second Order tensors"""

        # When a symmetric matrix is passed to the constructor
        mat = np.random.random((3,3))
        skew_sym_mat = mat - mat.T
        t = SkewSymmetricSecondOrderTensor(skew_sym_mat)
        np.testing.assert_array_equal(t.matrix, skew_sym_mat)

        # When an upper-diagonal matrix is passed to the constructor
        upper_mat = skew_sym_mat
        upper_mat[np.tril_indices(3)] = 0 # Set lower part to zero
        t2 = SkewSymmetricSecondOrderTensor(upper_mat)
        assert t == t2

        # Expect error in any other case
        expected_error = ('The input array must be either slices of symmetric matrices, of slices of upper-diagonal '
                          'matrices.')
        with self.assertRaises(ValueError) as context:
            _ = SymmetricSecondOrderTensor(mat)
        self.assertEqual(str(context.exception), expected_error)

    def test_add_sub_skew_symmetric_tensor(self):
        """Test basic operations between symmetric and skew-symmetric tensor, and check that the output classes are
        consistent."""
        mat = np.random.random((3,3))
        strain = StrainTensor(mat + mat.T)
        spin = SkewSymmetricSecondOrderTensor(mat - mat.T)

        # Check with consistent add.sub
        assert isinstance(strain + 2 * strain, StrainTensor)
        assert isinstance(spin + 2 * spin, SkewSymmetricSecondOrderTensor)

        # Check with inconsistent classes
        a = strain + 2 * spin
        assert (isinstance(a, SecondOrderTensor) and not isinstance(a, SymmetricSecondOrderTensor) and
                not isinstance(a, SkewSymmetricSecondOrderTensor))

        # Check negative values
        assert isinstance(-strain, StrainTensor)
        assert isinstance(-spin, SkewSymmetricSecondOrderTensor)

        # Check when adding a scaler value
        b = spin + 5
        assert (isinstance(b, SecondOrderTensor) and not isinstance(b, SymmetricSecondOrderTensor) and
                not isinstance(b, SkewSymmetricSecondOrderTensor))
        assert isinstance(strain + 5, StrainTensor)

        # Check when multiplying the tensor
        c = strain * spin
        assert (isinstance(c, SecondOrderTensor) and not isinstance(c, SymmetricSecondOrderTensor) and
                not isinstance(c, SkewSymmetricSecondOrderTensor))

        # Now check with rotations
        rotations = Rotation.random(100)
        assert isinstance(strain * rotations, StrainTensor)
        assert isinstance(spin * rotations, SkewSymmetricSecondOrderTensor)

    def test_to_pymatgen(self):
        """Test convertion from Elasticipy to pymatgen"""
        # First, try with a single tensor
        a = np.random.random((3, 3))
        a_sym = a + a.T
        strain = StrainTensor(a_sym)
        Strain_pymatgen = strain.to_pymatgen()
        np.testing.assert_array_equal(Strain_pymatgen.__array__(), a_sym)
        assert isinstance(Strain_pymatgen, mgStrain)

        # Now try with a 1D array
        n = 10
        b = np.random.random((n, 3, 3))
        b_sym = b + np.swapaxes(b, -1, -2)
        strain = StrainTensor(b_sym)
        Strain_pymatgen = strain.to_pymatgen()
        for i in range(n):
            np.testing.assert_array_equal(Strain_pymatgen[i].__array__(), b_sym[i])

        # Finally, try with a multidimensional array
        c = np.random.random((n, n, 3, 3))
        c_sym = c + np.swapaxes(c, -1, -2)
        strain = StrainTensor(c_sym)
        expected_error = 'The array must be flattened (1D tensor array) before converting to pytmatgen.'
        with self.assertRaises(ValueError) as context:
            _ = strain.to_pymatgen()
        self.assertEqual(str(context.exception), expected_error)

        # Now check that this also works for stress
        stress = StressTensor(a_sym)
        stress_pymatgen = stress.to_pymatgen()
        assert isinstance(stress_pymatgen, mgStress)

    def test_rand(self):
        """Test uniform random generation"""
        # Test two ways to define a rand tensor
        shape = (5,4)
        seed = 1324 # Ensure reproducibility
        t1 = SecondOrderTensor.rand(shape=shape, seed=seed)
        rng = np.random.default_rng(seed)
        t2 = SecondOrderTensor(rng.random(shape + (3,3)))
        assert np.all(t1==t2)

    def test_randn(self):
        """Test normal random distribution"""
        shape = (50, 40, 30)
        mean = np.random.random((3,3))
        std = np.random.random((3,3))
        t = SecondOrderTensor.randn(mean=mean, std=std, shape=shape)
        tmean = t.mean()
        tstd = t.std()
        tol = 1e-5
        np.testing.assert_array_almost_equal(tmean.matrix, mean, decimal=tol)
        np.testing.assert_array_almost_equal(tstd.matrix, std, decimal=tol)

    def test_inv(self):
        """Test inverse method"""
        shape = (3,4)
        matrix = np.random.random(shape + (3,3))
        t = SecondOrderTensor(matrix)
        tinv = t.inv()
        for i in range(shape[0]):
            for j in range(shape[1]):
                np.testing.assert_array_almost_equal(tinv[i,j].matrix, np.linalg.inv(matrix[i,j]))

    def test_tensile_shear(self):
        # Start with single tensile
        mag = 5
        t = Tensors.SymmetricSecondOrderTensor.tensile([1, 0, 0], mag)
        assert t.shape == ()
        mat = np.zeros((3, 3))
        mat[0, 0] = mag
        np.testing.assert_almost_equal(t.matrix, mat)

        # Now with arrays
        n = 5
        t = Tensors.SymmetricSecondOrderTensor.tensile([1,0,0], range(n))
        assert t.shape == (n,)
        mat = np.zeros((n,3,3))
        mat[:,0,0] = range(n)
        np.testing.assert_almost_equal(t.matrix, mat)

        # Try with a 2d directions
        with self.assertRaises(ValueError) as context:
            _ = Tensors.SymmetricSecondOrderTensor.tensile([1,0], range(n))
        self.assertEqual(str(context.exception), 'u must be 3D vector.')

        # Now shear
        t = Tensors.SymmetricSecondOrderTensor.shear([1,0,0], [0,1,0], range(n))
        assert t.shape == (n,)
        mat = np.zeros((n,3,3))
        mat[:,0,1] = mat[:,1,0] = range(n)
        np.testing.assert_almost_equal(t.matrix, mat)

    def test_repr(self):
        a=Tensors.StrainTensor.ones()
        assert a.__repr__() == 'Strain tensor\n[[1. 1. 1.]\n [1. 1. 1.]\n [1. 1. 1.]]'

        n=5
        b=Tensors.StressTensor.ones(n)
        assert b.__repr__() == 'Stress tensor\nShape=({},)'.format(n)

    def test_dot(self):
        m, n = 5, 6
        a_0d = StrainTensor.rand()
        a_1d = StrainTensor.rand((m,))
        a_2d = StressTensor.rand((m, n))
        b_0d = StrainTensor.rand()
        b_1d = StrainTensor.rand((m,))
        b_2d = StressTensor.rand((m, n))
        ab = a_0d.dot(b_0d)
        np.testing.assert_array_almost_equal(ab.matrix, np.matmul(a_0d.matrix, b_0d.matrix))
        ab = a_0d.dot(b_1d)
        for i in range(m):
            np.testing.assert_array_almost_equal(ab[i].matrix, np.matmul(a_0d.matrix, b_1d[i].matrix))
        ab = a_1d.dot(b_1d)
        for i in range(m):
            np.testing.assert_array_almost_equal(ab[i].matrix, np.matmul(a_1d[i].matrix, b_1d[i].matrix))
        ab = a_2d.dot(b_2d)
        for i in range(m):
            for j in range(n):
                np.testing.assert_array_almost_equal(ab[i,j].matrix, np.matmul(a_2d[i,j].matrix, b_2d[i,j].matrix))

        ab = a_1d.dot(b_1d, mode='cross')
        for i in range(m):
            for j in range(m):
                np.testing.assert_array_almost_equal(ab[i,j].matrix, np.matmul(a_1d[i].matrix, b_1d[j].matrix))
        ab = a_2d.dot(b_2d, mode='cross')
        for i in range(m):
            for j in range(n):
                for k in range(m):
                    for l in range(n):
                        np.testing.assert_array_almost_equal(ab[i,j,k,l].matrix, np.matmul(a_2d[i,j].matrix, b_2d[k,l].matrix))


    def test_rotate_orix(self):
        m, n = 5, 6
        t_0d = StrainTensor.rand()
        t_1d = StrainTensor.rand((m,))
        t_2d = StressTensor.rand((m, n))
        g_0d = OrixRot.random()
        g_1d = OrixRot.random(m)
        g_2d = OrixRot.random((m, n))

        a_rot = t_0d.rotate(g_0d)
        g_mat = g_0d.to_matrix()[0]
        np.testing.assert_array_almost_equal(a_rot.matrix, np.matmul(np.matmul(g_mat, t_0d.matrix), g_mat.T), )

        a_rot = t_1d.rotate(g_0d)
        for i in range(m):
            g_mat = g_0d.to_matrix()[0]
            np.testing.assert_array_almost_equal(a_rot[i].matrix, np.matmul(np.matmul(g_mat, t_1d[i].matrix), g_mat.T), )

        a_rot = t_1d.rotate(g_1d)
        for i in range(m):
            g_mat = g_1d.to_matrix()[i]
            np.testing.assert_array_almost_equal(a_rot[i].matrix, np.matmul(np.matmul(g_mat, t_1d[i].matrix), g_mat.T), )

        a_rot = t_0d.rotate(g_1d)
        for i in range(m):
            g_mat = g_1d.to_matrix()[i]
            np.testing.assert_array_almost_equal(a_rot[i].matrix, np.matmul(np.matmul(g_mat, t_0d.matrix), g_mat.T), )

        a_rot = t_2d.rotate(g_2d)
        for i in range(m):
            for j in range(n):
                g_mat = g_2d.to_matrix()[i,j]
                np.testing.assert_array_almost_equal(a_rot[i,j].matrix, np.matmul(np.matmul(g_mat, t_2d[i,j].matrix), g_mat.T), )

        # Try with 'cross' option
        a_rot = t_1d.rotate(g_1d, mode='cross')
        assert a_rot.shape == (m,m)
        for i in range(m):
            for j in range(m):
                g_mat = g_1d.to_matrix()[j]
                np.testing.assert_array_almost_equal(a_rot[i,j].matrix, np.matmul(np.matmul(g_mat, t_1d[i].matrix), g_mat.T), )

        a_rot = t_1d.rotate(g_2d, mode='cross')
        assert a_rot.shape == (m,m,n)
        for i in range(m):
            for j in range(m):
                for k in range(m):
                    g_mat = g_2d.to_matrix()[j,k]
                    np.testing.assert_array_almost_equal(a_rot[i,j,k].matrix, np.matmul(np.matmul(g_mat, t_1d[i].matrix), g_mat.T), )

    def test_stress_Voigt(self):
        m, n = 3, 2
        shape = (m, n)
        stress = StressTensor.rand(shape)
        stress_voigt = stress.to_Voigt()
        assert stress_voigt.shape == shape + (6,)
        for i in range(m):
            for j in range(n):
                assert stress_voigt[i, j, 0] == stress.matrix[i, j, 0, 0]
                assert stress_voigt[i, j, 1] == stress.matrix[i, j, 1, 1]
                assert stress_voigt[i, j, 2] == stress.matrix[i, j, 2, 2]
                assert stress_voigt[i, j, 3] == stress.matrix[i, j, 2, 1]
                assert stress_voigt[i, j, 4] == stress.matrix[i, j, 2, 0]
                assert stress_voigt[i, j, 5] == stress.matrix[i, j, 1, 0]
        assert np.all(stress == StressTensor.from_Voigt(stress_voigt))

    def test_strain_Voigt(self):
        m, n = 3, 2
        shape = (m, n)
        strain = StrainTensor.rand(shape)
        strain_voigt = strain.to_Voigt()
        assert strain_voigt.shape == shape + (6,)
        for i in range(m):
            for j in range(n):
                assert strain_voigt[i, j, 0] == strain.matrix[i, j, 0, 0]
                assert strain_voigt[i, j, 1] == strain.matrix[i, j, 1, 1]
                assert strain_voigt[i, j, 2] == strain.matrix[i, j, 2, 2]
                assert strain_voigt[i, j, 3] == 2 * strain.matrix[i, j, 2, 1]
                assert strain_voigt[i, j, 4] == 2 * strain.matrix[i, j, 2, 0]
                assert strain_voigt[i, j, 5] == 2 * strain.matrix[i, j, 1, 0]
        assert np.all(strain == StrainTensor.from_Voigt(strain_voigt))

    def test_stress_Kelvin(self):
        m, n = 3, 2
        shape = (m, n)
        stress = StressTensor.rand(shape)
        stress_kelvin = stress.to_Kelvin()
        assert stress_kelvin.shape == shape + (6,)
        s = np.sqrt(2)
        for i in range(m):
            for j in range(n):
                assert stress_kelvin[i, j, 0] == stress.matrix[i, j, 0, 0]
                assert stress_kelvin[i, j, 1] == stress.matrix[i, j, 1, 1]
                assert stress_kelvin[i, j, 2] == stress.matrix[i, j, 2, 2]
                assert stress_kelvin[i, j, 3] == s * stress.matrix[i, j, 2, 1]
                assert stress_kelvin[i, j, 4] == s * stress.matrix[i, j, 2, 0]
                assert stress_kelvin[i, j, 5] == s * stress.matrix[i, j, 1, 0]
        np.testing.assert_array_almost_equal(stress.matrix, StressTensor.from_Kelvin(stress_kelvin).matrix)

    def test_strain_Kelvin(self):
        m, n = 3, 2
        shape = (m, n)
        strain = StrainTensor.rand(shape)
        strain_kelvin = strain.to_Kelvin()
        assert strain_kelvin.shape == shape + (6,)
        s = np.sqrt(2)
        for i in range(m):
            for j in range(n):
                assert strain_kelvin[i, j, 0] == strain.matrix[i, j, 0, 0]
                assert strain_kelvin[i, j, 1] == strain.matrix[i, j, 1, 1]
                assert strain_kelvin[i, j, 2] == strain.matrix[i, j, 2, 2]
                assert strain_kelvin[i, j, 3] == s * strain.matrix[i, j, 2, 1]
                assert strain_kelvin[i, j, 4] == s * strain.matrix[i, j, 2, 0]
                assert strain_kelvin[i, j, 5] == s * strain.matrix[i, j, 1, 0]
        np.testing.assert_array_almost_equal(strain.matrix, StrainTensor.from_Kelvin(strain_kelvin).matrix)

    def test_deprecated_path(self):
        expected_warn = ("The module 'Elasticipy.StressStrainTensors' is deprecated and will be removed in a future "
                            "release. Please use 'Elasticipy.tensors.stress_strain' instead.")
        with self.assertWarns(DeprecationWarning) as context:
            from Elasticipy.StressStrainTensors import StressTensor, StrainTensor
        self.assertEqual(str(context.warning), expected_warn)

    def test_stack(self):
        size = 5
        a = SecondOrderTensor.rand(shape=size)
        b = SecondOrderTensor.rand(shape=size)
        c = SecondOrderTensor.stack((a, b))
        assert c.shape == (2, size)
        assert np.all(c[0] == a) and np.all(c[1] == b)
        c2 = SecondOrderTensor.stack((a,b), axis=1)
        assert c2.shape == (size, 2)
        assert np.all(c2[:,0] == a) and np.all(c2[:,1] == b)
        c3 = SecondOrderTensor.stack((a,b), axis=-1)
        assert np.all(c3 == c2)

    def test_Mohr_circles(self):
        s11 = 5
        t = StressTensor.tensile([1,0,0], s11)
        fig, ax = t.draw_Mohr_circles()
        assert ax.get_xlabel() == 'Normal stress'
        assert ax.get_ylabel() == 'Shear stress'
        assert np.all(ax.get_xticks() == (0, 0.5*s11, s11))
        assert np.all(ax.get_yticks() == (-0.5*s11, 0., 0.5*s11))

        e12 = 5
        t = StrainTensor.shear([1,0,0], [0,1,0], e12)
        fig, ax = t.draw_Mohr_circles()
        assert ax.get_xlabel() == 'Normal strain'
        assert ax.get_ylabel() == 'Shear strain'
        assert np.all(ax.get_xticks() == (-e12, -e12/2, 0, e12/2, e12))
        assert np.all(ax.get_yticks() == (-e12, -e12/2, 0, e12/2, e12))

if __name__ == '__main__':
    unittest.main()



import warnings
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.spatial.transform import Rotation
ALPHABET = 'abcdefghijklmnopqrstuv'

class _MatrixProxy:
    def __init__(self, matrix):
        self.matrix = matrix

    def __getitem__(self, args):
        sub = self.matrix[(...,) + (args if isinstance(args, tuple) else (args,))]
        if sub.shape == ():
            return float(sub)
        else:
            return sub

    def __setitem__(self, args, value):
        self.matrix[(...,) + (args if isinstance(args, tuple) else (args,))] = value

def _tensor_from_direction_magnitude(u, v, magnitude):
    if np.asarray(u).shape != (3,):
        raise ValueError('u must be 3D vector.')
    if np.asarray(v).shape != (3,):
        raise ValueError('v must be 3D vector.')
    u = u / np.linalg.norm(u)
    v = v / np.linalg.norm(v)
    direction_matrix = np.outer(u, v)
    if np.asarray(magnitude).ndim:
        return np.einsum('ij,...p->...pij', direction_matrix, magnitude)
    else:
        return magnitude * direction_matrix

def _transpose_matrix(matrix):
    return np.swapaxes(matrix, -1, -2)

def _symmetric_part(matrix):
    return 0.5 * (matrix + _transpose_matrix(matrix))

def _orientation_shape(g):
    if is_orix_rotation(g):
        return g.shape
    else:
        return (len(g),)

def _is_single_rotation(rotation):
    if isinstance(rotation, Rotation):
        return rotation.single
    elif is_orix_rotation(rotation):
        return rotation.size == 1
    else:
        raise TypeError('The input argument must be of class scipy.transform.Rotation or '
                        'orix.quaternion.rotation.Rotation')

_voigt_numbering = [[0, 0], [1, 1], [2, 2], [1, 2], [0, 2], [0, 1]]

def _unmap(array, mapping_convention):
    array = np.asarray(array)
    shape = array.shape
    if shape and (shape[-1] == 6):
        new_shape = shape[:-1] + (3, 3)
        unmapped_matrix = np.zeros(new_shape)
        for i in range(6):
            unmapped_matrix[..., _voigt_numbering[i][0], _voigt_numbering[i][1]] = array[..., i] / mapping_convention[i]
        return unmapped_matrix
    else:
        raise ValueError("array must be of shape (6,) or (...,6) with Voigt vector")

def _map(matrix, mapping_convention):
    shape = matrix.shape[:-2] + (6,)
    array = np.zeros(shape)
    for i in range(6):
        j, k = _voigt_numbering[i]
        array[...,i] = matrix[...,j,k]
    return array * mapping_convention

def filldraw_circle(ax, center, radius, color, fill=False, alpha=1.):
    theta = np.linspace(0, 2 * np.pi, 500)
    x = center[0] + radius * np.cos(theta)
    y = center[1] + radius * np.sin(theta)
    if fill:
        ax.fill(x, y, color=color, alpha=alpha)
    else:
        ax.plot(x, y, color=color)

kelvin_mapping = [1, 1, 1, np.sqrt(2), np.sqrt(2), np.sqrt(2)]

class SecondOrderTensor:
    """
    Template class for manipulation of second order tensors or arrays of second order tensors

    Attributes
    ----------
    matrix : np.ndarray
        (...,3,3) matrix storing all the components of the tensor

    """
    name = 'Second-order tensor'
    "Name to use when printing the tensor"

    def __init__(self, matrix):
        """
        Create an array of second-order tensors.

        The input argument can be:
            - an array of shape (3,3) defining all the components of the tensor;
            - a stack of matrices, that is an array of shape (...,3,3).

        Parameters
        ----------
        matrix : list or np.ndarray
            (3,3) matrix, stack of (3,3) matrices
        """
        matrix = np.array(matrix)
        shape = matrix.shape
        if len(shape) > 1 and shape[-2:] == (3, 3):
            self.matrix = matrix
        else:
            raise ValueError('The input matrix must be of shape (3,3) or (...,3,3)')

    def __repr__(self):
        s = self.name + '\n'
        if self.shape:
            s += 'Shape={}'.format(self.shape)
        else:
            s += self.matrix.__str__()
        return s

    def __getitem__(self, index):
        return self.__class__(self.matrix[index])

    def __setitem__(self, index, value):
        if isinstance(value, (float, np.ndarray)):
            self.matrix[index] = value
        elif type(value) == self.__class__:
            self.matrix[index] = value.matrix
        else:
            raise NotImplementedError('The r.h.s must be either float, a ndarray or an object of class {}'.format(self.__class__))

    def __add__(self, other):
        if type(self) == type(other):
            return self.__class__(self.matrix + other.matrix)
        elif isinstance(other, (int, float, np.ndarray)):
            mat = self.matrix + other
            if isinstance(self, SkewSymmetricSecondOrderTensor):
                return SecondOrderTensor(mat)
            else:
                return self.__class__(mat)
        elif isinstance(other, SecondOrderTensor):
            return SecondOrderTensor(self.matrix + other.matrix)
        else:
            raise NotImplementedError('The element to add must be a number, a numpy.ndarray or a tensor.')

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        if type(self) == type(other):
            return self.__class__(self.matrix - other.matrix)
        elif isinstance(other, (int, float, np.ndarray)):
            return self.__class__(self.matrix - other)
        else:
            raise NotImplementedError('The element to subtract must be a number, a numpy ndarray or a tensor.')

    def __neg__(self):
        return self.__class__(-self.matrix)

    def __rsub__(self, other):
        return -self + other

    @property
    def shape(self):
        """
        Return the shape of the tensor array

        Returns
        -------
        tuple
            Shape of array

        See Also
        --------
        ndim : number of dimensions
        """
        *shape, _, _ = self.matrix.shape
        return tuple(shape)

    @property
    def ndim(self):
        """
        Return the number of dimensions of the tensor array

        Returns
        -------
        int
            number of dimensions

        See Also
        --------
        shape : shape of tensor array
        """
        return len(self.shape)

    @property
    def C(self):
        """
        Return tensor components

        For instance T.C[i,j] returns all the (i,j)-th components of each tensor in the array.

        Returns
        -------
        np.ndarray
            Tensor components
        """
        return _MatrixProxy(self.matrix)

    def eig(self):
        """
        Compute the eigenvalues and eigenvectors of the tensor

        Returns
        -------
        lambda : np.ndarray
            Eigenvalues of each tensor.
        v : np.ndarray
            Eigenvectors of teach tensor.

        See Also
        --------
        eigvals : return only the eigenvalues (without directions)
        principal_directions : return only the principal directions (without eigenvalues)
        """
        return np.linalg.eig(self.matrix)

    def eigvals(self):
        """
        Compute the eigenvalues of the tensor, without computing the associated eigenvectors

        Returns
        -------
        numpy.ndarray
            Eigenvalues

        See Also
        --------
        eig : compute the eigenvalues and the eigenvector
        """
        return np.linalg.eigvals(self.matrix)

    def principal_directions(self):
        """
        Principal directions of the tensors

        Returns
        -------
        np.ndarray
            Principal directions of each tensor of the tensor array

        See Also
        --------
        eig : Return both eigenvalues and corresponding principal directions
        """
        return self.eig()[1]

    @property
    def I1(self):
        """
        First invariant of the tensor (trace)

        Returns
        -------
        np.ndarray or float
            First invariant(s) of the tensor(s)

        See Also
        --------
        I2 : Second invariant of the tensors
        I3 : Third invariant of the tensors (det)
        """
        return self.matrix.trace(axis1=-1, axis2=-2)

    @property
    def I2(self):
        """
        Second invariant of the tensor

        For a matrix M, it is defined as::

            I_2 = 0.5 * ( np.trace(M)**2 + np.trace(np.matmul(M, M.T)) )

        Returns
        -------
        np.array or float
            Second invariant(s) of the tensor(s)

        See Also
        --------
        I1 : First invariant of the tensors (trace)
        I3 : Third invariant of the tensors (det)
        """
        a = self.I1**2
        b = np.matmul(self.matrix, self._transpose_tensor()).trace(axis1=-1, axis2=-2)
        return 0.5 * (a - b)

    @property
    def I3(self):
        """
        Third invariant of the tensor (determinant)

        Returns
        -------
        np.array or float
            Third invariant(s) of the tensor(s)

        See Also
        --------
        I1 : First invariant of the tensors (trace)
        I2 : Second invariant of the tensors
        """
        return np.linalg.det(self.matrix)

    @property
    def J1(self):
        """
        First invariant of the deviatoric part of the stress tensor. It is always zeros, as the deviatoric part as null
        trace.

        Returns
        -------
        float or np.ndarray
            zero(s)
        """
        if self.shape:
            return np.zeros(self.shape)
        else:
            return 0.0

    @property
    def J2(self):
        """
        Second invariant of the deviatoric part of the tensor.

        Returns
        -------
        float or np.ndarray
            J2 invariant
        """
        return -self.deviatoric_part().I2

    @property
    def J3(self):
        """
        Third invariant of the deviatoric part of the tensor.

        Returns
        -------
        float or np.ndarray
            J3 invariant
        """
        return self.deviatoric_part().I3

    def Lode_angle(self, degrees=False):
        """
        Computes the Lode angle of the tensor.

        The returned value is defined from the positive cosine (see Notes).

        Parameters
        ----------
        degrees : bool, optional
            Whether to return the angle in degrees or not

        Returns
        -------
        float or numpy.ndarray

        See Also
        --------
        J2 : Second invariant of the deviatoric part
        J3 : Third invariant of the deviatoric part

        Notes
        -----
        The Lode angle is defined such that:

        .. math::

            \\cos(3\\theta)= \\frac{J_3}{2}\\left(\\frac{3}{J_2}\\right)^{3/2}
        """
        J2 = np.atleast_1d(self.J2)
        J3 = np.atleast_1d(self.J3)
        non_hydro =  J2 !=0.
        cosine = np.ones(shape=J3.shape) * np.nan
        cosine[non_hydro] = J3[non_hydro] / 2 * (3 / J2[non_hydro] )**(3 / 2)
        if degrees:
            theta = np.arccos(cosine) * 60 / np.pi
        else:
            theta = np.arccos(cosine) / 3
        if self.shape:
            return theta
        else:
            return theta[0]

    def trace(self):
        """
        Return the traces of the tensor array

        Returns
        -------
        np.ndarray or float
            traces of each tensor of the tensor array

        See Also
        --------
        I1 : First invariant of the tensors (trace)
        I2 : Second invariant of the tensors
        I3 : Third invariant of the tensors (det)
        """
        return self.I1

    def __mul__(self, B):
        """
        Element-wise matrix multiplication of arrays of tensors. Each tensor of the resulting tensor array is computed
        as the matrix product of the tensor components.

        Parameters
        ----------
        B : SecondOrderTensor or np.ndarray or Rotation or float
            If B is a numpy array, we must have::

                B.shape == (..., 3, 3)

        Returns
        -------
            Array of tensors populated with element-wise matrix multiplication.

        See Also
        --------
        matmul : matrix-like multiplication of tensor arrays
        """
        if isinstance(B, SecondOrderTensor):
            return self.dot(B, mode='pair')
        elif isinstance(B, Rotation) or is_orix_rotation(B):
            return self.rotate(B, mode='pair')
        elif isinstance(B, (float, int)):
            return self.__class__(self.matrix * B)
        elif isinstance(B, np.ndarray):
            if B.shape == self.shape:
                new_matrix = np.einsum('...ij,...->...ij', self.matrix, B)
                return self.__class__(new_matrix)
            elif B.shape == self.matrix.shape:
                return self.__class__(np.matmul(self.matrix, B))
            else:
                err_msg = 'For a tensor of shape {}, the input argument must be an array of shape {} or {}'.format(
                    self.shape, self.shape, self.shape + (3, 3))
                raise ValueError(err_msg)
        else:
            raise ValueError('The input argument must be a tensor, an ndarray, a rotation or a scalar value.')

    def rotate(self, rotation, mode='pair'):
        """
        Apply rotation(s) to the tensor(s).

        The rotations can be applied element-wise, or on each cross-combination (see below).

        Parameters
        ----------
        rotation : scipy.spatial.Rotation or orix.quaternion.Rotation
        mode : str, optional
            If 'pair', the rotations are applied element wise. Broadcasting rule applies.
            If 'cross', all the possible combinations are considered. If ``C=A.rotate(rot)``, then
            ``C.shape==A.shape + rot.shape``.

        Returns
        -------
        SecondOrderTensor
        """
        if self.shape == ():
            ein_str = '...li,...kj,lk->...ij'
        elif _is_single_rotation(rotation):
            ein_str = 'li,kj,...lk->...ij'
        else:
            if mode=='pair':
                ein_str = '...li,...kj,...lk->...ij'
            elif mode=='cross':
                ndim_0 = self.ndim
                ndim_1 = len(_orientation_shape(rotation))
                indices_self = ALPHABET[:ndim_0]
                indices_g = ALPHABET[:ndim_1].upper()
                indices_res = indices_self + indices_g
                ein_str = indices_g + 'zw,' + indices_g + 'yx,' + indices_self + 'zy->' + indices_res + 'wx'
            else:
                raise ValueError('Invalid mode. It can be "cross" or "pair".')
        g_mat = rotation_to_matrix(rotation)
        matrix = np.einsum(ein_str, g_mat, g_mat, self.matrix)
        return self.__class__(matrix)

    def __rmul__(self, other):
        if isinstance(other, (float, int)):
            return self.__mul__(other)
        else:
            raise NotImplementedError('Left multiplication is only implemented for scalar values.')

    def __truediv__(self, other):
        new_mat = np.zeros(self.matrix.shape)
        non_zero = np.any(self.matrix, axis=(-1, -2))
        if isinstance(other, (float, int)):
            new_mat[non_zero] = self.matrix[non_zero] / other # Hack to force 0/0 = 0
        elif isinstance(other, np.ndarray) and (self.shape == other.shape):
            new_mat[non_zero] = np.einsum('pij,p->pij', self.matrix[non_zero], 1/other[non_zero])
            return self.__class__(new_mat)
        else:
            raise NotImplementedError('Tensors can only be divided by scalar values or by arrays of the same shape.')
        return self.__class__(new_mat)

    def __eq__(self, other):
        """
        Check whether the tensors in the tensor array are equal

        Parameters
        ----------
        other : SecondOrderTensor or np.ndarray
            Tensor to compare with

        Returns
        -------
        numpy.ndarray
            True element is True if the corresponding tensors are equal.
        """
        if isinstance(other, SecondOrderTensor):
            return self == other.matrix
        elif isinstance(other, np.ndarray):
            if (other.shape == (3,3)) or (other.shape == self.shape + (3,3)):
                return np.all(self.matrix == other, axis=(-2, -1))
            else:
                raise ValueError('The value to compare must be an array of shape {} or {}'.format(self.shape, self.shape + (3,3)))

    def dot(self, other, mode='pair'):
        """
        Perform contraction product ("dot product") between tensor.

        On tensor arrays, the product contraction can be performed element-wise, or considering all cross-combinations
        (see below).

        Parameters
        ----------
        other : SecondOrderTensor
            tensor or tensor array to compute the product from
        mode : str, optional
            If 'pair' (default), the contraction products of tensor arrays are applied element-wise. Broadcasting rule
             applies.

            If 'cross', all combinations of contraction product are considered. If ``C=A.dot(B,mode='cross')``, then
            ``C.shape==A.shape + B.shape``

        Returns
        -------
        SecondOrderTensor

        Examples
        --------
        >>> from Elasticipy.tensors.second_order import SecondOrderTensor
        >>> A=SecondOrderTensor.rand(10)
        >>> B=SecondOrderTensor.rand(10)
        >>> AB_pair = A.dot(B)
        >>> AB_pair.shape
        (10,)

        >>> AB_cross = A.dot(B, mode='cross')
        >>> AB_cross.shape
        (10, 10)

        We can for instance check that:

        >>> AB_pair[5] == A[5].dot(B[5])
        True

        and:

        >>> AB_cross[0,1] == A[0].dot(B[1])
        True

        See Also
        --------
        ddot : Double-contraction product
        """
        if self.shape == ():
            ein_str = 'ik,...kj->...ij'
        else:
            if mode=='pair':
                ein_str = '...ik,...kj->...ij'
            elif mode=='cross':
                ndim_0 = self.ndim
                ndim_1 = other.ndim
                indices_0 = ALPHABET[:ndim_0]
                indices_1 = ALPHABET[:ndim_1].upper()
                indices_2 = indices_0 + indices_1
                ein_str = indices_0 + 'ik,' + indices_1 + 'kj->' + indices_2 + 'ij'
            else:
                raise ValueError('Invalid mode. Use "pair" or "cross".')
        matrix = np.einsum(ein_str, self.matrix, other.matrix)
        return SecondOrderTensor(matrix)


    def matmul(self, other):
        """
        Perform matrix-like product between tensor arrays. Each "product" is a matrix product between
        the tensor components.

        If A.shape=(a1, a2, ..., an) and B.shape=(b1, b2, ..., bn), with C=A.matmul(B), we have::

            C.shape = (a1, a2, ..., an, b1, b2, ..., bn)

        and::

            C[i,j,k,...,p,q,r...] = np.matmul(A[i,j,k,...], B[p,q,r,...])

        Parameters
        ----------
        other : SecondOrderTensor or np.ndarray or Rotation
            Tensor array or rotation to right-multiply by. If Rotation is provided, the rotations are applied on each
            tensor.

        Returns
        -------
        SecondOrderTensor
            Tensor array

        See Also
        --------
        __mul__ : Element-wise matrix product
        """
        warnings.warn(
            'matmul() is deprecated and will be removed in a future version. Use dot(tensor,mode="cross") or '
            'rotate(rotation,mode="cross") instead.',
            DeprecationWarning,
            stacklevel=2)
        if isinstance(other, SecondOrderTensor):
            return self.dot(other, mode='cross')
        elif isinstance(other, Rotation) or is_orix_rotation(Rotation):
            return self.rotate(other, mode='cross')
        else:
            raise ValueError('The input argument must be either a rotation or a SecondOrderTensor')

    def transpose_array(self):
        """
        Transpose the array of tensors

        If A is a tensor array of shape [s1, s2, ..., sn], A.T is of shape [sn, ..., s2, s1].

        Returns
        -------
        SecondOrderTensor
            Transposed array

        See Also
        --------
        T : transpose the tensor array (not the components)
        """
        if self.ndim < 2:
            return self
        else:
            matrix = self.matrix
            ndim = matrix.ndim
            new_axes = np.hstack((ndim - 3 - np.arange(ndim - 2), -2, -1))
            transposed_arr = np.transpose(matrix, new_axes)
            return self.__class__(transposed_arr)

    @property
    def T(self):
        """
        Transpose the array of tensors.

        It is actually an alias for transpose_array()

        Returns
        -------
        SecondOrderTensor
            Transposed array
        """
        return self.transpose_array()

    def _transpose_tensor(self):
        return _transpose_matrix(self.matrix)

    def transpose_tensor(self):
        """
        Transpose of tensors of the tensor array

        Returns
        -------
        SecondOrderTensor
            Array of transposed tensors of the tensor array

        See Also
        --------
        transpose_array : transpose the array (not the components)
        """
        return self.__class__(self._transpose_tensor())

    def ddot(self, other, mode='pair'):
        """
        Double dot product (contraction of tensor product, usually denoted ":") of two tensors.

        For two tensors whose matrices are M1 and M2::

            M1.ddot(M2) == np.trace(np.matmul(M1, M2))

        Parameters
        ----------
        other : SecondOrderTensor or np.ndarray
            Tensor or tensor array to multiply by before contraction.
        mode : str, optional
            If "pair", the dot products are performed element-wise before contraction. Broadcasting rule applies.
            If "cross", all the cross-combinations are computed, increasing the dimensionality.
            If ``C=A.ddot(B, mode='cross')``, then ``C.shape = A.shape + B.shape``.


        Returns
        -------
        float or np.ndarray
            Result of double dot product

        See Also
        --------
        dot : contraction product ("dot product") between tensor.

        """
        tensor_prod = self.transpose_tensor().dot(other, mode=mode)
        return tensor_prod.trace()

    def _flatten(self):
        if self.shape:
            new_len = np.prod(self.shape)
            return np.reshape(self.matrix, (new_len, 3, 3))
        else:
            return self.matrix

    def _stats(self, fun, axis=None):
        if axis is None:
            flat_mat = self._flatten()
            new_matrix = fun(flat_mat, axis=0)
        else:
            if axis < 0:
                axis += -2
            if (axis > self.ndim - 1) or (axis < -self.ndim - 2):
                raise ValueError('The axis index is out of bounds for tensor array of shape {}'.format(self.shape))
            new_matrix = fun(self.matrix, axis=axis)
        return self.__class__(new_matrix)

    def flatten(self):
        """
        Flatten the array of tensors.

        If T is of shape [s1, s2, ..., sn], the shape for T.flatten() is [s1*s2*...*sn].

        Returns
        -------
        SecondOrderTensor
            Flattened array (vector) of tensor

        See Also
        --------
        ndim : number of dimensions of the tensor array
        shape : shape of the tensor array
        reshape : reshape a tensor array
        """
        return self.__class__(self._flatten())

    def reshape(self, shape, **kwargs):
        """
        Reshape the array of tensors

        Parameters
        ----------
        shape : tuple
            New shape of the array
        kwargs : dict
            Keyword arguments passed to numpy.reshape()

        Returns
        -------
        SecondOrderTensor
            Reshaped array

        See Also
        --------
        flatten : flatten an array to 1D
        """
        new_matrix = self.matrix.reshape(shape + (3,3,), **kwargs)
        return self.__class__(new_matrix)

    def mean(self, axis=None):
        """
        Arithmetic mean value

        Parameters
        ----------
        axis : int or None, default None
            Axis to compute the mean along with.
            If None, returns the overall mean (mean of flattened array)

        Returns
        -------
        SecondOrderTensor
            Mean tensor

        See Also
        --------
        std : Standard deviation
        min : Minimum value
        max : Maximum value
        """
        if self.ndim:
            return self._stats(np.mean, axis=axis)
        else:
            return self

    def std(self, axis=None):
        """
        Standard deviation

        Parameters
        ----------
        axis : int or None, default None
            Axis to compute standard deviation along with.
            If None, returns the overall standard deviation (std of flattened array)

        Returns
        -------
        SecondOrderTensor
            Tensor of standard deviation

        See Also
        --------
        mean : Mean value
        min : Minimum value
        max : Maximum value
          """
        if self.ndim:
            return self._stats(np.std, axis=axis)
        else:
            return self.__class__(np.zeros((3, 3)))

    def min(self, axis=None):
        """
        Minimum value

        Parameters
        ----------
        axis : int or None, default None
           Axis to compute minimum along with.
           If None, returns the overall minimum (min of flattened array)

        Returns
        -------
        SecondOrderTensor
           Minimum value of tensors

        See Also
        --------
        max : Maximum value
        mean : Mean value
        std : Standard deviation
        """
        if self.ndim:
            return self._stats(np.min, axis=axis)
        else:
            return self

    def max(self, axis=None):
        """
        Maximum value

        Parameters
        ----------
        axis : int or None, default None
            Axis to compute maximum along with.
            If None, returns the overall maximum (max of flattened array)

        Returns
        -------
        SecondOrderTensor
            Maximum value of tensors

        See Also
        --------
        min : Minimum value
        mean : Mean value
        std : Standard deviation
        """
        if self.ndim:
            return self._stats(np.max, axis=axis)
        else:
            return self

    def _symmetric_part(self):
        return 0.5 * (self.matrix + self._transpose_tensor())

    def symmetric_part(self):
        """
        Symmetric part of the tensor

        Returns
        -------
        SymmetricSecondOrderTensor
            Symmetric tensor

        See Also
        --------
        skewPart : Skew-symmetric part of the tensor
        """
        return SymmetricSecondOrderTensor(self._symmetric_part())

    def skew_part(self):
        """
        Skew-symmetric part of the tensor

        Returns
        -------
        SkewSymmetricSecondOrderTensor
            Skew-symmetric tensor
        """
        new_mat = 0.5 * (self.matrix - self._transpose_tensor())
        return SkewSymmetricSecondOrderTensor(new_mat)

    def spherical_part(self):
        """
        Spherical (hydrostatic) part of the tensor

        Returns
        -------
        self
            Spherical part

        See Also
        --------
        I1 : compute the first invariant of the tensor
        deviatoricPart : deviatoric the part of the tensor
        """
        s = self.I1 / 3
        return self.eye(self.shape)*s

    def deviatoric_part(self):
        """
        Deviatoric part of the tensor

        Returns
        -------
        self

        See Also
        --------
        sphericalPart : spherical part of the tensor
        """
        return self - self.spherical_part()

    @classmethod
    def eye(cls, shape=()):
        """
        Create an array of tensors populated with identity matrices

        Parameters
        ----------
        shape : tuple or int, default ()
            If not provided, it just creates a single identity tensor. Otherwise, the tensor array will be of the
            specified shape.

        Returns
        -------
        cls
            Array of identity tensors

        See Also
        --------
        ones : creates an array of tensors full of ones
        zeros : creates an array full of zero tensors
        """
        if isinstance(shape, int):
            matrix_shape = (shape, 3, 3)
        else:
            matrix_shape = shape + (3, 3,)
        eye = np.zeros(matrix_shape)
        eye[..., np.arange(3), np.arange(3)] = 1
        return cls(eye)

    @classmethod
    def ones(cls, shape=()):
        """
        Create an array of tensors populated with matrices of full of ones.

        Parameters
        ----------
        shape : tuple or int, default ()
            If not provided, it just creates a single tensor of ones. Otherwise, the tensor array will be of the
            specified shape.

        Returns
        -------
        cls
            Array of ones tensors

        See Also
        --------
        eye : creates an array of identity tensors
        zeros : creates an array full of zero tensors
        """
        if isinstance(shape, int):
            matrix_shape = (shape, 3, 3)
        else:
            matrix_shape = shape + (3, 3,)
        ones = np.ones(matrix_shape)
        return cls(ones)

    @classmethod
    def zeros(cls, shape=()):
        """
        Create an array of tensors populated with matrices full of zeros.

        Parameters
        ----------
        shape : tuple or int, default ()
            If not provided, it just creates a single tensor of ones. Otherwise, the tensor array will be of the
            specified shape.

        Returns
        -------
        cls
            Array of ones tensors

        See Also
        --------
        eye : creates an array of identity tensors
        ones : creates an array of tensors full of ones
        """
        if isinstance(shape, int):
            matrix_shape = (shape, 3, 3)
        else:
            matrix_shape = shape + (3, 3,)
        zeros = np.zeros(matrix_shape)
        return cls(zeros)

    @classmethod
    def tensile(cls, u, magnitude):
        """
        Create an array of tensors corresponding to tensile state along a given direction.

        Parameters
        ----------
        u : np.ndarray or list
            Tensile direction. Must be a 3D vector.
        magnitude : float or np.ndarray or list
            Magnitude of the tensile state to consider. If a list or an array is provided, the shape of the tensor array
            will be of the same shape as magnitude.
        Returns
        -------
        cls
            tensor or tensor array
        """
        mat = _tensor_from_direction_magnitude(u, u, magnitude)
        return cls(mat)

    @classmethod
    def rand(cls, shape=None, seed=None):
        """
        Generate a tensor array, populated with random uniform values in [0,1).

        Parameters
        ----------
        shape : int or tuple, optional
            Shape of the tensor array. If not provided, a single tensor is returned
        seed : int, optional
            Sets the seed for random generation. Useful to ensure reproducibility

        Returns
        -------
        cls
            Tensor or tensor array of uniform random value

        See Also
        --------
        randn : Generate a random sample of tensors whose components follows a normal distribution

        Examples
        --------
        Generate a single random tensor:

        >>> from Elasticipy.tensors.second_order import SecondOrderTensor as tensor
        >>> tensor.rand(seed=123)
        Second-order tensor
        [[0.68235186 0.05382102 0.22035987]
         [0.18437181 0.1759059  0.81209451]
         [0.923345   0.2765744  0.81975456]]

        Now try with tensor array:
        >>> t = tensor.rand(shape=(100,50))
        >>> t.shape
        (100,50)
        """
        if shape is None:
            shape = (3,3)
        elif isinstance(shape, int):
            shape = (shape, 3, 3)
        else:
            shape = shape + (3,3)
        rng = np.random.default_rng(seed)
        a = rng.random(shape)
        if issubclass(cls, SymmetricSecondOrderTensor):
            a = _symmetric_part(a)
        return cls(a)

    def inv(self):
        """Compute the reciprocal (inverse) tensor"""
        return SecondOrderTensor(np.linalg.inv(self.matrix))

    @classmethod
    def randn(cls, mean=np.zeros((3,3)), std=np.ones((3,3)), shape=None, seed=None):
        """
        Generate a tensor array, populated with components follow a normal distribution.

        Parameters
        ----------
        mean : list of numpy.ndarray, optional
            (3,3) matrix providing the mean values of the components.
        std : list of numpy.ndarray, optional
            (3,3) matrix providing the standard deviations of the components.
        shape : tuple, optional
            Shape of the tensor array
        seed : int, optional
            Sets the seed for random generation. Useful to ensure reproducibility

        Returns
        -------
        cls
            Tensor or tensor array of normal random value
        """
        if shape is None:
            new_shape = (3,3)
        else:
            new_shape = shape + (3,3)
        rng = np.random.default_rng(seed)
        mat = np.zeros(new_shape)
        mean = np.asarray(mean)
        std = np.asarray(std)
        for i in range(0,3):
            for j in range(0,3):
                mat[...,i,j] = rng.normal(mean[i,j], std[i,j], shape)
        if issubclass(cls, SymmetricSecondOrderTensor):
            mat = _symmetric_part(mat)
        return cls(mat)

    @classmethod
    def shear(cls, u, v, magnitude):
        """
        Create an array of tensors corresponding to shear state along two orthogonal directions.

        Parameters
        ----------
        u : np.ndarray or list
            First direction. Must be a 3D vector.
        v : np.ndarray or list
            Second direction. Must be a 3D vector.
        magnitude : float or np.ndarray or list
            Magnitude of the shear state to consider. If a list or an array is provided, the shape of the tensor array
            will be of the same shape as magnitude.
        Returns
        -------
        cls
            tensor or tensor array
        """
        if np.abs(np.dot(u, v)) > 1e-5:
            raise ValueError("u and v must be orthogonal")
        mat = _tensor_from_direction_magnitude(u, v, magnitude)
        return cls(mat)

    def div(self, axes=None, spacing=1.):
        """
        Compute the divergence vector of the tensor array, along given axes.

        If the tensor has n dimensions, the divergence vector will be computed along its m first axes, with
        m = min(n, 3), except if specified in the ``axes`` parameter (see below).

        Parameters
        ----------
        axes : list of int, tuple of int, int or None, default None
            Indices of axes along which to compute the divergence vector. If None (default), the m first axes of the
            array will be used to compute the derivatives.
        spacing : float or np.ndarray or list, default 1.
            Spacing between samples the in each direction. If a scalar value is provided, the spacing is assumed equal
            in each direction. If an array or a list is provided, spacing[i] must return the spacing along the i-th
            axis (spacing[i] can be float or np.ndarray).

        Returns
        -------
        np.ndarray
            Divergence vector of the tensor array. If the tensor array is of shape (m,n,...,q), the divergence vector
            will be of shape (m,n,...,q,3).

        Notes
        -----
        The divergence of a tensor field :math:`\\mathbf{t}(\\mathbf{x})` is defined as:

        .. math::

            [\\nabla\\cdot\\mathbf{t}]_i = \\frac{\\partial t_{ij}}{\\partial x_j}

        The main application of this operator is for balance of linear momentum of stress tensor:

        .. math::

            \\rho \\mathbf{\\gamma} = \\nabla\\cdot\\mathbf{\\sigma} + \\rho\\mathbf{b}

        where :math:`\\mathbf{\\sigma}` is the stress tensor, :math:`\\mathbf{\\gamma}` is the acceleration,
        :math:`\\mathbf{b}` is the body force density and :math:`\\rho` is the mass density.

        In this function, the derivatives are computed with ``numpy.grad`` function.
        """
        ndim = min(self.ndim, 3)    # Even if the array has more than 3Ds, we restrict to 3D
        if isinstance(spacing, (float, int)):
            spacing = [spacing, spacing, spacing]
        if axes is None:
            axes = range(ndim)
        elif isinstance(axes, int):
            axes = (axes,)
        elif not isinstance(axes, (tuple, list)):
            raise TypeError("axes must be int, tuple of int, or list of int.")
        if len(axes) > ndim:
            error_msg = ("The number of axes must be less or equal to the number of dimensions ({}), "
                         "and cannot exceed 3").format(self.ndim)
            raise ValueError(error_msg)
        else:
            ndim = len(axes)
        if max(axes) >= ndim:
            raise IndexError("axes index must be in range of dimensions ({})".format(self.ndim))
        div = np.zeros(self.shape + (3,))
        for dim in range(0, ndim):
            div += np.gradient(self.C[:,dim], spacing[dim], axis=axes[dim])
        return div

    def save(self, file, **kwargs):
        """
        Save the tensor array as binary file (.npy format).

        This function uses numpy.save function.

        Parameters
        ----------
        file : file, str or pathlib.Path
            File or filename to which the tensor is saved.
        kwargs : dict
            Keyword arguments passed to numpy.save()

        See Also
        --------
        load_from_npy : load a tensor array from a numpy file
        """
        np.save(file, self.matrix, **kwargs)

    @classmethod
    def load_from_npy(cls, file, **kwargs):
        """
        Load a tensor array for .npy file.

        This function uses numpy.load()

        Parameters
        ----------
        file : file, str or pathlib.Path
            File to read to create the array
        kwargs : dict
            Keyword arguments passed to numpy.load()

        Returns
        -------
        SecondOrderTensor
            Tensor array

        See Also
        --------
        save : save the tensor array as a numpy file
        """
        matrix = np.load(file, **kwargs)
        if matrix.shape[-2:] != (3,3):
            raise ValueError('The shape of the array to load must be (...,3,3).')
        else:
            return cls(matrix)

    def save_as_txt(self, file, name_prefix='', **kwargs):
        """
        Save the tensor array to human-readable text file.

        The array must be 1D. The i-th row of the file will provide the components of the i-th tensor in of the array.
        This function uses pandas.DataFrame.to_csv().

        Parameters
        ----------
        file : file or str
            File to dump tensor components to.
        name_prefix : str, optional
            Prefix to add for naming the columns. For instance, name_prefix='E' will result in columns named E11, E12,
            E13 etc.
        kwargs : dict
            Keyword arguments passed to pandas.DataFrame.to_csv()
        """
        if self.ndim > 1:
            raise ValueError('The array must be flatten before getting dumped to text file.')
        else:
            d = dict()
            for i in range(3):
                if isinstance(self, SkewSymmetricSecondOrderTensor):
                    r = range(i+1, 3)   # If the tensor is skew-symmetric, there is no need to save the full matrix
                elif isinstance(self, SymmetricSecondOrderTensor):
                    r = range(i, 3)     # Idem, except that we also need the diagonal
                else:
                    r =range(3)
                for j in r:
                    key = name_prefix + '{}{}'.format(i+1, j+1)
                    d[key] = self.C[i,j]
            df = pd.DataFrame(d)
            df.to_csv(file, index=False, **kwargs)

    @classmethod
    def load_from_txt(cls, file, name_prefix='', **kwargs):
        """
        Load a tensor array from text file.

        Parameters
        ----------
        file : str or file
            Textfile to read the components from.
        name_prefix : str, optional
            Prefix to add to each column when parsing the file. For instance, with name_prefix='E', the function will
            look for columns names E11, E12, E13 etc.

        Returns
        -------
        SecondOrderTensor
            Flat (1D) tensor constructed from the values given in the text file
        """
        df = pd.read_csv(file, **kwargs)
        matrix = np.zeros((len(df), 3, 3))
        for i in range(3):
            if cls is SkewSymmetricSecondOrderTensor:
                r = range(i+1, 3)
            elif cls is SymmetricSecondOrderTensor:
                r = range(i, 3)
            else:
                r= range(3)
            for j in r:
                key = name_prefix + '{}{}'.format(i + 1, j + 1)
                matrix[:, i, j] = df[key]
        return cls(matrix)

    def to_pymatgen(self):
        """
        Convert the second order object into an object compatible with pymatgen.

        The object to use must be either a single tensor, or a flat tensor array. In the latter case, the output will be
        a list of pymatgen's tensors.

        Returns
        -------
        pymatgen.analysis.elasticity.Strain, pymatgen.analysis.elasticity.Stress, pymatgen.core.tensors.Tensor or list
            The type of output depends on the type of object to use:
                - if the object is of class StrainTensor, the output will be of class pymatgen.analysis.elasticity.Strain
                - if the object is of class StressTensor, the output will be of class pymatgen.analysis.elasticity.Stress
                - otherwise, the output will be of class pymatgen.core.tensors.Tensor

        See Also
        --------
        flatten : Converts a tensor array to 1D tensor array
        """
        try:
            from Elasticipy.tensors.stress_strain import StrainTensor, StressTensor
            if isinstance(self, StrainTensor):
                from pymatgen.analysis.elasticity import Strain as Constructor
            elif isinstance(self, StressTensor):
                from pymatgen.analysis.elasticity import Stress as Constructor
            else:
                from pymatgen.core.tensors import Tensor as Constructor
        except ImportError:
            raise ModuleNotFoundError('Module pymatgen is required for this function.')
        if self.ndim > 1:
            raise ValueError('The array must be flattened (1D tensor array) before converting to pytmatgen.')
        if self.shape:
            return [Constructor(self[i].matrix) for i in range(self.shape[0])]
        else:
            return Constructor(self.matrix)

    @classmethod
    def stack(cls, arrays, axis=0):
        """
        Stack tensor arrays along the specified axis.

        Parameters
        ----------
        arrays : list of SecondOrderTensor or tuple of SecondOrderTensor
            List of tensor to stack together
        axis : int, optional
            Axis along which to stack

        Returns
        -------
        SecondOrderTensor
            Stacked tensor array

        Examples
        --------
        >>> from Elasticipy.tensors.second_order import SecondOrderTensor
        >>> import numpy as np
        >>> a = SecondOrderTensor.rand(shape=3)
        >>> b = SecondOrderTensor.rand(shape=3)
        >>> c = SecondOrderTensor.stack((a, b))
        >>> c.shape
        (2, 3)
        >>> np.all(c[0] == a)
        True
        >>> np.all(c[1] == b)
        True

        >>> a = SecondOrderTensor.rand(shape=(3, 4))
        >>> b = SecondOrderTensor.rand(shape=(3, 4))
        >>> c = SecondOrderTensor.stack((a, b), axis=1)
        >>> c.shape
        (3, 2, 4)
        >>> np.all(c[:,0,:] == a)
        True
        >>> np.all(c[:,1,:] == b)
        True
        """
        mat_array = [a.matrix for a in arrays]
        if axis<0:
            axis = axis - 2
        mat_stacked = np.stack(mat_array, axis=axis)
        return cls(mat_stacked)


class SymmetricSecondOrderTensor(SecondOrderTensor):
    voigt_map = [1, 1, 1, 1, 1, 1]
    "List of factors to use for building a tensor from Voigt vector(s)"

    name = 'Symmetric second-order tensor'

    def __init__(self, mat, force_symmetry=False):
        """
        Create a symmetric second-order tensor

        Parameters
        ----------
        mat : list or numpy.ndarray
            matrix or array to construct the symmetric tensor. It must be symmetric with respect to the two last indices
            (mat[...,i,j]=mat[...,j,i]), or composed of slices of upper-diagonal matrices (mat[i,j]=0 for each i>j).
        force_symmetry : bool, optional
            If true, the symmetric part of the matrix will be used. It is mainly meant for debugging purpose.

        Examples
        --------
        We can create a symmetric tensor by privoding the full matrix, as long it is symmetric:

        >>> from Elasticipy.tensors.second_order import SymmetricSecondOrderTensor
        >>> a = SymmetricSecondOrderTensor([[11, 12, 13],[12, 22, 23],[13, 23, 33]])
        >>> print(a)
        Symmetric second-order tensor
        [[11. 12. 13.]
         [12. 22. 23.]
         [13. 23. 33.]]


        Alternatively, we can pass the upper-diagonal part only:

        >>> b = SymmetricSecondOrderTensor([[11, 12, 13],[0, 22, 23],[0, 0, 33]])

        and check that a==b:

        >>> a==b
        True
        """
        if isinstance(mat, SecondOrderTensor):
            mat = mat.matrix
        else:
            mat = np.asarray(mat, dtype=float)
        mat_transposed = _transpose_matrix(mat)
        if np.all(np.isclose(mat, mat_transposed)) or force_symmetry:
            # The input matrix is symmetric
            super().__init__(0.5 * (mat + mat_transposed))
        elif np.all(mat[..., np.tril_indices(3, k=-1)[0], np.tril_indices(3, k=-1)[1]] == 0):
            # The input matrix is upper-diagonal
            lower_diagonal = np.zeros_like(mat)
            triu_indices = np.triu_indices(3,1)
            lower_diagonal[..., triu_indices[0], triu_indices[1]] = mat[..., triu_indices[0], triu_indices[1]]
            super().__init__(mat + _transpose_matrix(lower_diagonal))
        else:
            raise ValueError('The input array must be either slices of symmetric matrices, of slices of upper-diagonal '
                             'matrices.')

    @classmethod
    def from_Voigt(cls, array, voigt_map=None):
        """
        Construct a SymmetricSecondOrderTensor from a Voigt vector, or slices of Voigt vectors.

        If the array is of shape (6,), a single tensor is returned. If the array is of shape (m,n,o,...,6), the tensor
        will be of shape (m,n,o,...).

        Parameters
        ----------
        array : np.ndarray or list
            array to build the SymmetricSecondOrderTensor from. We must have array.ndim>0 and array.shape[-1]==6.

        voigt_map : list or tuple, optional
            6-lenght list of factors to use for mapping. If None (default), the default Voigt map of the constructor is
            used.

        Returns
        -------
        SymmetricSecondOrderTensor

        See Also
        --------
        from_Kelvin : Construct a tensor from vector(s) following the Kelvin notation

        Examples
        --------
        >>> from Elasticipy.tensors.second_order import SymmetricSecondOrderTensor
        >>> SymmetricSecondOrderTensor.from_Voigt([11, 22, 33, 23, 13, 12])
        Symmetric second-order tensor
        [[11. 12. 13.]
         [12. 22. 23.]
         [13. 23. 33.]]
        """
        if voigt_map is None:
            voigt_map = cls.voigt_map
        matrix = _unmap(array, voigt_map)
        return cls(matrix)

    def to_Voigt(self):
        """
        Convert the tensor to vector, or slices of vector, following the Voigt convention.
        
        If the tensor array has shape (m,n,...), the result will be of shape (m,n,...,6).
        
        Returns
        -------
        numpy.ndarray
            Voigt vector summarizing the components
        """
        return _map(self.matrix, self.voigt_map)
    
    @classmethod
    def from_Kelvin(cls, array):
        """
        Build a tensor from the Kelvin vector, or slices of Kelvin vectors

        Parameters
        ----------
        array : np.ndarray or list
            Vectors, or slices of vectors, consisting in components following the Kelvin convention
        Returns
        -------
        SymmetricSecondOrderTensor

        See Also
        --------
        from_Voigt : construct a tensor from vector(s) following the Voigt notation
        to_Kelvin : convert the tensor to vector(s) following the Kelvin convention
        """
        matrix = _unmap(array, kelvin_mapping)
        return cls(matrix)
    
    def to_Kelvin(self):
        """
        Convert the tensor to vector, or slices of vector, following the Kelvin(-Mandel) convention.

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        from_Kelvin : Construct a tensor from vector(s) following the Kelvin convention
        to_Voigt : Convert the tensor to vector(s) following the Voigt convention
        """
        return _map(self.matrix, kelvin_mapping)

    def eig(self):
        """
        Compute the principal values (eigenvalues) and principal direction (eigenvectors) of the tensor, sorted in
        descending order of principal values

        Returns
        -------
        numpy.ndarray
            Principal values
        numpy.ndarray
            Principal directions

        See Also
        --------
        eigvals : compute the principal values only
        """
        eigvals, eigdir = np.linalg.eigh(self.matrix)
        return eigvals[..., ::-1], eigdir[..., :, ::-1]

    def eigvals(self):
        """
        Compute the principal values (eigenvalues), sorted in descending order.

        Returns
        -------
        numpy.ndarray
            Principal values

        See Also
        --------
        eig : return the principal values and principal directions
        """
        eigvals = np.linalg.eigvalsh(self.matrix)
        return np.flip(eigvals,axis=-1)

    def draw_Mohr_circles(self):
        """
        Draw the Mohr circles of the symmetric second-order tensor

        Given a tensor, the Mohr circles are meant to visually illustrate the possible shear components one can find
        when randomly rotating the tensor. See `here <https://en.wikipedia.org/wiki/Mohr%27s_circle>`_ for details.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The Matplotlib figure containing the plot.
        ax : matplotlib.axes._axes.Axes
            The Matplotlib axes of the plot.
        """
        c,b,a = self.eigvals()

        # Sizes and locations of circles
        r1 = (c - b) / 2
        r2 = (b - a) / 2
        r3 = (c - a) / 2
        center1 = ((b + c) /2, 0)
        center2 = ((a + b) /2, 0)
        center3 = ((a + c) /2, 0)

        fig, ax = plt.subplots()
        filldraw_circle(ax, center1, r1, 'skyblue')
        filldraw_circle(ax, center2, r2, 'lightgreen')
        filldraw_circle(ax, center3, r3, 'red')
        filldraw_circle(ax, center3, r3, 'red', fill=True, alpha=0.2)
        filldraw_circle(ax, center1, r1, 'white', fill=True)
        filldraw_circle(ax, center2, r2, 'white', fill=True)
        ax.set_aspect('equal')
        ax.set_xlabel(f"Normal")
        ax.set_ylabel(f"Shear")
        ax.grid(True)
        xticks = (a,b,c, center1[0], center2[0])
        ax.set_xticks(np.unique(xticks))
        yticks = (-r3, -r2, -r1, 0, r1, r2, r3)
        ax.set_yticks(np.unique(yticks))

        return fig, ax


class SkewSymmetricSecondOrderTensor(SecondOrderTensor):
    name = 'Skew-symmetric second-order tensor'

    def __init__(self, mat, force_skew_symmetry=False):
        """Class constructor for skew-symmetric second-order tensors

        Parameters
        ----------
        mat : list or numpy.ndarray
            Input matrix, or slices of matrices. Each matrix should be skew-symmetric, or have zero-component on lower -
            diagonal part (including the diagonal).

        Examples
        --------
        One can construct a skew-symmetric tensor by providing the full skew-symmetric matrix:

        >>> from Elasticipy.tensors.second_order import SkewSymmetricSecondOrderTensor
        >>> a = SkewSymmetricSecondOrderTensor([[0, 12, 13],[-12, 0, 23],[-13, -23, 0]])
        >>> print(a)
        Skew-symmetric second-order tensor
        [[  0.  12.  13.]
         [-12.   0.  23.]
         [-13. -23.   0.]]

        Alternatively, one can pass the upper-diagonal part only:

        >>> b = SkewSymmetricSecondOrderTensor([[0, 12, 13],[0, 0, 23],[0, 0, 0]])

        and check that a==b:

        >>> a==b
        True

        """
        mat = np.asarray(mat, dtype=float)
        mat_transposed = _transpose_matrix(mat)
        if np.all(np.isclose(mat, -mat_transposed)) or force_skew_symmetry:
            # The input matrix is symmetric
            super().__init__(0.5 * (mat - mat_transposed))
        elif np.all(mat[..., np.tril_indices(3)[0], np.tril_indices(3)[1]] == 0):
            # The input matrix is upper-diagonal
            super().__init__(mat - mat_transposed)
        else:
            raise ValueError('The input array must be either slices of skew-symmetric matrices, of slices of upper-'
                             'diagonal matrices with zero-diagonal.')


def rotation_to_matrix(rotation, return_transpose=False):
    """
    Converts a rotation to slices of matrices

    Parameters
    ----------
    rotation : scipy.spatial.Rotation or orix.quaternion.Rotation
        Object to convert
    return_transpose : bool, optional
        If true, it will also return the transpose matrix as a 2nd output argument
    Returns
    -------
    numpy.ndarray or tuple
        Rotation matrices
    """
    if isinstance(rotation, Rotation):
        matrix = rotation.as_matrix()
    elif is_orix_rotation(rotation):
        inv_rotation = ~rotation
        matrix = inv_rotation.to_matrix()
        if matrix.shape == (1,3,3):
            matrix = matrix[0]
    else:
        raise TypeError('The input argument must be of class scipy.transform.Rotation or '
                        'orix.quaternion.rotation.Rotation')
    if return_transpose:
        return matrix, _transpose_matrix(matrix)
    else:
        return matrix


def is_orix_rotation(other):
    """
    Check whether the argument is a rotation from Orix by looking at the existing methods.

    Parameters
    ----------
    other : any
        object to test
    Returns
    -------
    bool
        True if other.to_matrix() exists
    """
    return hasattr(other, "to_matrix") and callable(getattr(other, "to_matrix"))



import numpy as np
from elastic import Elastic
from Elasticipy.tensors.elasticity import StiffnessTensor
import matplotlib as mpl
mpl.use('Qt5Agg')   # Ensure interactive plot
from Elasticipy.spherical_function import sph2cart, _plot3D
import time
import matplotlib.pyplot as plt


C = StiffnessTensor.monoclinic(phase_name='TiNi',
                               C11=231, C12=127, C13=104,
                               C22=240, C23=131, C33=175,
                               C44=81, C55=11, C66=85,
                               C15=-18, C25=1, C35=-3, C46=3)
Celate = Elastic(list(C._matrix))

# Plotting with Elate
start_time=time.perf_counter()
n_phi, n_theta = 500,500
phi = np.linspace(0, 2*np.pi, n_phi)
theta = np.linspace(0, np.pi, n_theta)

u= np.zeros((n_phi, n_theta, 3))
Eelate = np.zeros((n_phi, n_theta))
for i in range(n_phi):
    for j in range(n_theta):
        Eelate[i,j] = Celate.Young([theta[j], phi[i]])
        u[i,j,:] = sph2cart(phi[i], theta[j])
fig = plt.figure()
ax = _plot3D(fig, u.reshape(n_phi, n_theta, 3), Eelate)
t = time.perf_counter() - start_time
ax.set_title("Elate (t={:.3f} s)".format(t))

# Plotting with Elasticpy
start_time=time.perf_counter()
fig, ax = C.Young_modulus.plot3D(n_phi=n_phi, n_theta=n_theta)
t = time.perf_counter() - start_time
#ax.set_title("Elasticipy (t={:.3f} s)".format(t))
fig.savefig('../JOSS/Plot_E.png', dpi=300)
